﻿/* --- FMfindRepl_Ancillary.jsx ------ UTF-8 ---------------------------------------------------
              Array functions, string functions, debugging aids etc.
Comment       to be included in FMfindRepl.jsx
History       2020-12-29
*/
KLD_F.BackupSettings = function () { // ===========================================================
/*            Backup the settings file - it may be modified in this session
Returns       Function value is xxx
Called by     PalFindRepl
Calling       Message
Comment       The settings file must be located in the same directory as the script
              The backup file will get the name xxx_backup.xml
Reference     xxx
Used in       FMfindRepl
History       2020-12-14
*/
var e, fBackup, fXmlFile, sXmlFile, sBackup;

  sXmlFile = KLD_F.APP_NAME + ".xml";             // FMfindRepl.xml
  sBackup  = KLD_F.APP_NAME + "_backup.xml";
  fXmlFile = File($.fileName.replace (/[^\\\/]+$/i , sXmlFile));
  fBackup  = new File($.fileName.replace (/[^\\\/]+$/i , sBackup));

  try {
      fXmlFile.copy(fBackup);
      return  true;
  } catch (e) {                                   // what kind of error can this be ?
      KLD_F.Message ("E", KLD_F.UItxt.BackupSettings.@Msg_01, "BackupSettings", sXmlFile, e);
      fXmlFile.close();
      return false;
  }
} //--- end BackupSettings ------------------------------------------------------------------------

KLD_F.CompareString = function (str1, str2, sSortOrder) { // ======================================
/*            compare function for SortLocal
Arguments     str1, str2      Strings to be compared
              sSortOrder      String with the characters in the desired order. See calling routine
Returns       0 for equality, 1 for str1 < str2, 2 for str1 > str2
Called by     SortLocal
Calling       -
Reference     Klaus Göbel
History       2020-11-23 test with VSCode and ESDebugger
*/
var j, iPosInStr1, iPosInStr2, k, result = 0;

  k = str1.length;
  for (j = 0; j <= k; j++) {
    if (str1.charAt(j) != str2.charAt(j)) {
      iPosInStr1 = sSortOrder.indexOf(str1.charAt(j));
      iPosInStr2 = sSortOrder.indexOf(str2.charAt(j));
      if (iPosInStr1 > iPosInStr2) {
        result = 1;
      } else {
        result = 2;
      }
    break;
    }
  }
  return result;
} //--- end CompareString -------------------------------------------------------------------------

KLD_F.ConvertSelfClosingElements = function (xFile) { // ==========================================
/*            Expand self-terminating XML elements to empty start-tag end-tag elements
Arguments     xFile           XML file to be modified
Returns       false in case of error, else true
Called by     ButtonDelete, ButtonSave, ButtonSaveNew
Calling       Message, Replacer1
Comment       Self-terminating XML elements create problems when they are read. The buttons
              Retrieve, Save, SaveNew, Delete and Reset do not work any more.
Reference     https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/replace
Used in       FMfindRepl
History       2021-03-22
*/
var BOM = "\uFEFF";

  if ((xFile == null) || (xFile == undefined)) {
    return false;
  } else if (xFile.indexOf("/") > 0 || xFile.indexOf("\\") > 0) { // fully qualified path
    fXFile = new File(xFile);
  } else {
    fXFile = new File($.fileName.replace (/[^\\\/]+$/i , xFile));  // file in script folder
  }
  if (fXFile.exists === false) {
    KLD_F.Message ("E", KLD_F.UItxt.ConvertSelfClosingElements.@msg_01, "ConvertSelfClosingElements", sXmlFile, e.message);
      return false;
  }
  fXFile.open("r");
  xData = fXFile.read();
  fXFile.close ();
  xData = xData.replace (/<(\w+)\/>/g , KLD_F.Replacer1);
  fXFile.open('w');
  fXFile.encoding = "UTF-8";
  fXFile.write(BOM);
  fXFile.write(xData);
  fXFile.close();

  return true;
} //--- end ConvertSelfClosingElements ---------------------------------------------------------

KLD_F.CopyToClipboard = function (theString) { //==================================================
/*            Copy string to clipboard
Arguments     theString       String which may contain several lines - does this really work?
Returns       Function value is xxx
Called by     SettingsToPasteBuffer
Calling       -
Attention     NL must be inserted as String.fromCharCode(10);
              TAB must be inserted as String.fromCharCode(08);
Comment       A textframe with a paragraph on the first masterpage (which always exists) is crated,
              becasue FM gets copy only from inside...
              To avoid flicker and jump to the fist page, some measures are taken
Reference     Based on Rick Quatro's method
Used in       FMbiblio, FMfindRepl
History       2021-02-08
*/
var oDoc, masterPage, oLocation = {}, oPgf, textFrame, textLoc, tRange, PT = 65536;

  oDoc = KLD_F.oCurrentDoc;
  if (app.Displaying == true) {                   // avoid flicker
    app.Displaying = false;
    app.ActiveDoc.Redisplay();                    // FM is very lousy in refreshing
  }
  oLocation.pgf    = oDoc.TextSelection.beg.obj;  // save current location
  oLocation.txtRange = oDoc.TextSelection;
  oLocation.page   = oDoc.CurrentPage;
// --- start of the core function ----------------
  tRange = new TextRange;
  masterPage = oDoc.FirstMasterPageInDoc;         // Get first master page in document.
  textFrame = oDoc.NewTextFrame (masterPage.PageFrame); // Add a temporary text frame
  textFrame.Width  = 480 * PT;
  oPgf = textFrame.FirstPgf;
  textLoc = new TextLoc (oPgf, 0);
  oDoc.AddText (textLoc, theString);              // write string into this paragraph

  tRange.beg.obj = oPgf;                          // select this paragraph
  tRange.beg.offset = 0;
  tRange.end.obj = oPgf;
  tRange.end.offset = theString.length;
  oDoc.TextSelection = tRange;

  oDoc.Copy (0);                                  // copy selection to clipboard not interactive
  textFrame.Delete ();                            // delete the text frame
// --- end of the core function ------------------
  oDoc.CurrentPage = oLocation.page;              // restore current location
  oPgf = oLocation.pgf;
  if (oPgf.ObjectValid () === 1) {
    oDoc.TextSelection =   oLocation.txtRange;
  }
  oDoc.ScrollToText(oLocation.txtRange);
  if (app.Displaying == false) {                  // restore previous state
    app.Displaying = true;
    app.ActiveDoc.Redisplay();
  }
} // --- end CopyToClipboard ----------------------------------------------------------------------

KLD_F.DisplayObject = function (obj) { // =========================================================
/*            Make an object visible by selection
Arguments     obj             Object to be displayed. A graphic object will be selected.
Returns       -
Called by     FindColourUse, FindOverflowCell, FindOverflowText
Calling       IsGraphicObject, GetPage, GetCellRowNum
Attention     Not checked for other than graphic objects, Cell and Pfg
Used in       FMfindRepl
History       2021-02-25
*/
var jCol, jRow, oDoc = KLD_F.oCurrentDoc, oPage, oTblProps={}, oTLbeg, oTLend, oTR;

  if (KLD_F.IsGraphicObject(obj)) {               // OK
    oPage = KLD_F.GetPage (oDoc, obj);
    oDoc.CurrentPage = oPage;
    obj.GraphicIsSelected = true;
  } else if (obj.constructor.name == "Cell"){     // OK
    jRow    = KLD_F.GetCellRowNum (obj);
    jCol = obj.CellColNum;
    oRow    = obj.CellRow;
    oTbl    = oRow.RowTbl;
    oPage   = KLD_F.GetPage (oDoc, obj);
    oDoc.CurrentPage = oPage;
    oTbl.MakeTblSelection (jRow, jRow, jCol, jCol);
  } else if (obj.constructor.name == "Pgf"){      // OK
    oTLbeg = new TextLoc(obj,0);
    oTLend = new TextLoc(obj,Constants.FV_OBJ_END_OFFSET);
    oTR    = new TextRange(oTLbeg,oTLend);
    oDoc.TextSelection = oTR;
    oPage  = KLD_F.GetPage (oDoc, obj);
    oDoc.CurrentPage = oPage;
    oDoc.ScrollToText(oTR);
  }
} //--- end DisplayObject -------------------------------------------------------------------------

KLD_F.ExecuteFCodes = function (asCommands) { // ==================================================
/*            Execute keyboard commands, if nothing else helps
Arguments     asCommands      array of strings with command names, for example
              ["SelectRow", "ClearRemovingCells"]
Returns       -
Called by     GetSettingsList, PalFindRepl, GetCatalogAny, GetCatalogF
Calling       -
History       2020-12-17
*/
var j, nCommands, cmd, aFcodes = [];
  nCommands = asCommands.length;
  for (j = 0; j < nCommands; j++) {
    cmd = app.GetNamedCommand (asCommands[j]);
    Fcodes ([cmd.Fcode]);
  }
  return
} // --- end ExecuteFCode -------------------------------------------------------------------------

KLD_F.FillDlgList = function(oList, asListItems, bFirstBlank) { // ================================
/*            Fill dialogue list from global array
Arguments     oList           dialogue list to be filled from the array
              asListItems     array with the list items
              bFirstBlank     True: first item to be empty; False, not so
Returns       -
Called by     DebugMenu, GetSettingsList, PalFindRepl, GetCatalogAny, GetCatalogF, call-back functions
Calling       -
Attention     Clearing oList by the standard method for arrays tilts FM in the statement removeAll
              wPalF.g01.lstCatalgF.items.length = 0;       // this is invalid!
History       2020-11-10 l
*/
var j, nItems;

  nItems = asListItems.length;
  oList.removeAll ();                             // clear the list
  if (bFirstBlank) { oList.add ("item", "");}     // start with an empty entry
  for (j = 0; j < nItems; j++) {
    oList.add ("item", asListItems[j]);
  }
} //--- end FillDlgList ---------------------------------------------------------------------------

KLD_F.FindIndexFromValue = function (aSearchArray, sSearch) { //===================================
/*            Index of of particular item in array
Arguments     aSearchArray    array to be searched
              sSearch         may be part of what to find
Returns       Function value is Index or null
Called by     ButtonRestore, ButtonRetrieve, ButtonSave
Calling       -
Comment       Find at the beginning of the string
              Only first occurence in array element is searched
Reference     -
Used in       FMfindRepl
History       2020-11-10 l
*/
var j, sItem, saLen;

  saLen = aSearchArray.length;
  for (j = 0; j < saLen; j++) {
    if (aSearchArray[j].search(sSearch) == 0) {   // search must match at the string start
    return j;
    }
  }
  return null;
} //--- end FindIndexFromValue --------------------------------------------------------------------

KLD_F.GetCellRowNum = function  (oCell) { //=======================================================
/*            Get number of current table row
Arguments     oCell           The current cell object
Returns       Function value is current row within table (top heading row is 0)
Called by     DisplayObject
Calling       -
Reference     Ric and Jang, https://forums.adobe.com/thread/2229520
Used in       FMfindRepl
History       2021-03-01
*/
var iCounter = 0, oRow, oTbl;

  oCurrRow = oCell.CellRow;
  oTbl     = oCurrRow.RowTbl;
  oRow = oTbl.FirstRowInTbl;
  while (oRow.ObjectValid ()) {
    if (oRow.id === oCurrRow.id) {
      return iCounter;
    }
    iCounter += 1;
    oRow = oRow.NextRowInTbl;
  }
} //--- end GetCellRowNum -------------------------------------------------------------------------

KLD_F.GetClonedCase = function (sFound, sReplace) { // ============================================
/*            Get the replacement string with the cloned case of the found one
Arguments     sFound          Found string providing the case information
              sReplace        Replacement string as provided by users
Returns       Modified replacement string according to case information
Called by     ButtonReplace
Comment       Both strings must be tokenised, because the words may have different lengths.
              A token may end with a punctuation. Additional word in sReplace are not treated.
Reference     https://community.adobe.com/t5/framemaker-discussions/clone-case-of-replacement-string-in-regexp-sFound-replace-in-extendscript/td-p/13204487
Used in       FMfindRepl
History       2022-09-26
*/
var asFound, j, jMax, k, kMax, min, asReplace, sChar1, sChar2, sModified = "", sModWord;

  asFound = sFound.match(/\S+/g);                 // tokenise both strings
  asReplace = sReplace.match(/\S+/g);

  jMax = asFound.length;                          // number of tokens
  kMax = asReplace.length;
  min  = Math.min(jMax, kMax);
  for (j = 0; j < min; j++) {                     // corresponding words in either string
    sModWord = asReplace[j];
    for (k = 0; k < sModWord.length; k++) {
      sChar1 = asFound[j].substr(0,1);            // first char in word
      if (sChar1 == sChar1.toUpperCase ()) {
        sModWord = sModWord.charAt(0).toUpperCase () + sModWord.slice (1);
      } else {
        sModWord = sModWord.charAt (0).toLowerCase () + sModWord.slice (1);
      }
      if (asFound[j].length > 1 && asReplace[j].length > 1) {
        sChar2 = asFound[j].substr(1,1);          // second to last char in word
        if (sChar2 == sChar2.toUpperCase ()) {
          sModWord = sModWord.substr (0,1) + sModWord.substr(1).toUpperCase ()
        } else {
          sModWord = sModWord.substr (0,1) + sModWord.substr(1).toLowerCase ()
        }
      }
    }
    if (j > 0) {sModified = sModified + " "}
    sModified = sModified + sModWord;
  }
  if (j < kMax) {
    for (j = min; j < kMax; j++) {                // add the additional words
      sModified = sModified + " ";
      sModified = sModified + asReplace[j];
    }
  }
  return sModified;
} // --- end GetClonedCase ------------------------------------------------------------------------

KLD_F.GetColourNames = function () { // ===========================================================
/*            Fill the array asColourNames with localised System colour names from XML file
Returns       -
Called by     PalFindRepl
Calling       -
History       2021-02-23
*/
var j, n;

  KLD_F.asColourNames.length = 0;                 // clear array
  n = KLD_F.UItxt.asColourNames.item.length();
  for (j = 0; j < n; j++) {
    KLD_F.asColourNames.push (KLD_F.UItxt.asColourNames.item[j].toString());
  }
} //--- end of GetColourNames ---------------------------------------------------------------------

KLD_F.GetIniItemIndex = function (aArray, sSearch) { //============================================
/*            Get Index of of particular item in an ini array
Arguments     aArray          array of ini entries to be searched
              sSearch         may be part of what to find
Returns       Function value is Index of found item - or null
Called by     MakerIniRead
Calling       -
Comment       Only first occurence in array element is found
History       2020-12-29
*/
var j, saLen, searchLen, sItem;

  saLen     = aArray.length;
  searchLen = sSearch.length;
  for (j = 0; j < saLen; j++) {
    sItem = aArray[j].toUpperCase();
    if (sItem.indexOf(sSearch.toUpperCase()) != -1) {
    return j;
    }
  }
  return null;
} //--- end GetIniItemIndex -----------------------------------------------------------------------

KLD_F.GetDocOpenProps = function (bVisible) { //===================================================
/*            Set specific open property values to open the document or book
Arguments     bVisible        true: make document visible after open; false: invisible
Returns       Function value =  Open parameters for suppressing various open errors
Called by     OpenDocument
Calling       -
Comment       The list of openProps is similar to Rick's in SaveBookAsOldVersion.jsx
History       2020-12-29
*/
var j, openProps;

  openProps = GetOpenDefaultParams();
  j = GetPropIndex(openProps, Constants.FS_AlertUserAboutFailure);
    openProps[j].propVal.ival=false;              // no user-information/intervention
  j = GetPropIndex(openProps, Constants.FS_MakeVisible );
    openProps[j].propVal.ival=bVisible;           // open invisible
  j = GetPropIndex(openProps, Constants.FS_FileIsOldVersion);
    openProps[j].propVal.ival = Constants.FV_DoOK;// allow old version
  j = GetPropIndex(openProps, Constants.FS_FileIsInUse);
    openProps[j].propVal.ival = Constants.FV_ResetLockAndContinue; // unlock if necessary
  j = GetPropIndex(openProps, Constants.FS_FontNotFoundInDoc);
    openProps[j].propVal.ival = Constants.FV_DoOK;// allow missing fonts
  j=GetPropIndex(openProps, Constants.FS_FontChangedMetric);
    openProps[j].propVal.ival=Constants.FV_DoOK;  // allow changing fonts
  j=GetPropIndex(openProps, Constants.FS_FontNotFoundInCatalog);
    openProps[j].propVal.ival=Constants.FV_DoOK;  // unknown font ?
  j = GetPropIndex(openProps, Constants.FS_RefFileNotFound);
    openProps[j].propVal.ival = Constants.FV_AllowAllRefFilesUnFindable; // allow missing insets
  return (openProps);
} //--- end GetDocOpenProps -----------------------------------------------------------------------

KLD_F.GetSettingsList = function () { // ==========================================================
/*            Fill the UI settings-list from the XML file
Called by     PalFindRepl, ButtonDelete, ButtonSave, ButtonSaveNew
Calling       GetXMLdata, SortLocal, FillDlgList
History       2020-12-29
*/
var indexOfTAB, j, sName, asNames = [], nItems, sXmlFile;

  sXmlFile = KLD_F.APP_NAME + ".xml";             // FMfindRepl.xml
  KLD_F.asSettingName.length = 0;                 // clear array
  KLD_F.xSettings = KLD_F.GetXMLdata (sXmlFile);
  nItems = KLD_F.xSettings.item.length();
  for (j = 0; j < nItems; j++) {
    sName = KLD_F.xSettings.item[j].name.toString();
    KLD_F.asSettingName.push (sName + "\t" + j);  // append index of item in sXmlFile
  }
  KLD_F.asSettingName = KLD_F.SortLocal (KLD_F.asSettingName, KLD_F.APP_LANG);
  for (j = 0; j < nItems; j++) {                  // leave the orignal index out in the UI list
    indexOfTAB = KLD_F.asSettingName[j].indexOf("\t");
    asNames[j] = KLD_F.asSettingName[j].substr(0, indexOfTAB);
  }
  KLD_F.FillDlgList (wPalF.g09.lstSettingName, asNames, false);
} //--- end GetSettingsList -----------------------------------------------------------------------

KLD_F.GetSystemColourIndex = function (sColour) { // =============================================
/*            Get kReservedColor from the index within the list of colours
Arguments     sColour         Name of the colour as found in the list
Returns       Function value is 0 for user defined colour or n ≤ 16 for system colour
Called by     ButtonFind
Calling       IsInArray
Comment       The names of system colous are localised, hence can not be used with ease
History       2020-12-29
*/
var jColour;

  jColour = KLD_F.IsInArray(KLD_F.asColourNames, sColour);
  if (jColour == null) {
    return 0;                                     // User colour
  } else  {
    return jColour;                               // System colour
  }
} //--- end GetSystemColourIndex ------------------------------------------------------------------

KLD_F.GetText = function (oDoc, oText) { //=== Get the sText from the sText object ================
/*            Get the sText from the sText object
Arguments     oDoc            current document
              oText           object to get the string from
Returns       Function value is the text from the object
Called by     PreserveSpace, FindEmptyTextLine, FindInSelection, ButtonReplace
Calling       -
Reference     Rick Quatro, https://community.adobe.com/t5/framemaker/getting-text-from-footnotes/m-p/5177652
Used in       FMbiblio, FMcalc, FMfindRepl
History       2020-12-29
*/
var aTextItems, iTxtLen, j, sText= "";

  if (oText.constructor.name !== "TextRange") {   // Get a list of the strings in the Text object or Text range
    aTextItems = oText.GetText(Constants.FTI_String);  // text objects
  } else {
    aTextItems = oDoc.GetTextForRange(oText, Constants.FTI_String); // text range
  }
  iTxtLen = aTextItems.len;
  for (j = 0; j < iTxtLen; j += 1) {              // Concatenate the strings
    sText += (aTextItems[j].sdata);
  }
  return sText;                                   // Return the Text
} //--- end GetText -------------------------------------------------------------------------------

KLD_F.GetXMLdata = function (sXmlFile) { // =======================================================
/*            Read localisation strings from xml file
Argements     sXmlFile        fully qualified path of xml file or just the file name.
              File must exist
              If no filepath is given (no slash or backslash present),
              then the file is assumed in the same directory as the script
Returns       xml sructure
Called by     GetSettingsList, SwitchUIlanguageX
Calling       Message
Usage         xSettings = getXMLdata (sXmlFile);
Attention     this function is called by SwithUIlanguageX, hence localisation has not finished yet.
              Therefore: error messages need to be localised with the ExtendScript method
Reference     E:\FM-ExtendScript\SaveBookOldVersion\SaveBookAsOldVersion.jsx by Rick Quatro
History       2020-11-11
*/
var e, fXmlFile, xData, msg_01, msg_02;

  msg_01 = { en: "XML file «" + sXmlFile + "» not found.",
             de: "XML datei «" + sXmlFile + "» nicht gefunden.",
             fr: "Fichier XML «" + sXmlFile + "» non trouvé."};
  msg_02 = { en: "Read error on XML file «" + sXmlFile + "». Error:\n "+ e,
             de: "Lesefehler auf XML datei «" + sXmlFile + "» . Fehler:\n "+ e,
             fr: "Erreur de lecture sur le fichier XML «" + sXmlFile + "». Erreur:\n " + e};
  if ((sXmlFile == null) || (sXmlFile == undefined)) {
    return false;
  } else if (sXmlFile.indexOf("/") > 0 || sXmlFile.indexOf("\\") > 0) { // fully qualified path
    fXmlFile = File(sXmlFile);
  } else {
    fXmlFile = File($.fileName.replace (/[^\\\/]+$/i , sXmlFile));  // file in script folder
  }
  if (fXmlFile.exists === false) {
      KLD_F.Message ("E", localize (msg_01), "getXMLdata");
      return false;
  }
  fXmlFile.open("r");
  fXmlFile.encoding = "UTF-8";
  try {
      xData = new XML(fXmlFile.read());
      fXmlFile.close();
      return  xData;
  } catch (e) {                                   // what kind of error can this be ?
      KLD_F.Message ("E", localize (msg_02), "getXMLdata");
      fXmlFile.close();
      return false;
  }
} //--- end getXMLdata ----------------------------------------------------------------------------

KLD_F.IsInArray = function (array, item) { //======================================================
/*            Get index of item in array
Arguments     array           Array of items (strings, other)
              item            Item to be found in array
Returns       Function value is index or null
Called by     GetSystemColourIndex
Calling       -
Attention     Match of item type and arry item types not checked!
Used in       FMbibilo, FMcalc, FMvars, FMfindRepl
History       2018-09-26
*/
var j, jMax;

  jMax = array.length;
  for (j = 0; j < jMax; j++) {
    if (array [j] == item ) { return j;}
  }
  return null;                                    // not found in array
} //--- end IsInArray -----------------------------------------------------------------------------

KLD_F.IsEmptyString = function (string) { // ======================================================
/*            Is string really empty
Arguments     string          String to be tested
Returns       True or False
Called by     FindEmptyTextLine, ButtonSave, ButtonSaveNew
Calling       -
Used in       FMfindRepl
History       2021-02-09
*/
var re_Trim = /(^[\s\n\r\t\x0B]+)|([\s\n\r\t\x0B]+$)/g;
  string = string.replace(re_Trim, "");
  return string.length == 0;
} // end IsEmptyString ----------------------------------------------------------------------------

KLD_F.IsGraphicObject = function (obj) {
/*            Check whether the object belongs to the graphic objects
Arguments     obj              Object to be checked
Returns       true for graphic objects (see variable aGraphicObj) | false for other objects
Called by     DisplayObject, FindColourUse, GetNextObjectInDoc, ButtonFind
Calling       -
Comment       To be independant, IsInArray is integrated here
              The graphic object math must be written with lower case initital
              FDK lists FO_Math, ESTK lists Math, but an object inspection reveils it to be math
Used in       FMfindRepl
History       2021-02-25
*/
var aGraphicObj = ["AFrame", "Arc", "Ellipse", "Group", "Inset", "Line", "math", "MathML", "Polygon",
                   "Polyline", "Rectangle", "RoundRect", "TextFrame", "TextLine", "UnanchoredFrame"],
    j, jMax, sName;

  if (obj == undefined) return false;
  jMax  = aGraphicObj.length;
  sName = obj.constructor.name
  for (j = 0; j < jMax; j++) {
    if (aGraphicObj [j] == sName ) { return true;}
  }
  return false;
} //--- end IsGraphicObject -----------------------------------------------------------------------

KLD_F.IsGroupableObject = function (obj) {
/*            Check whether the object belongs to the graphic objects, which can be grouped
Arguments     obj              Object to be checked
Returns       true for groupabele objects (see variable aGraphicObj) | false for other objects
Called by     ButtonFind
Calling       -
Comment       The graphic object "math" must be written with lower case initital
              FDK lists FO_Math, ESTK lists Math, but an object inspection reveils it to be math
              Anchored and unanchored frames do not have property GroupParent
Used in       FMfindRepl
History       2022-05-18
*/
var aGraphicObj = ["Arc", "Ellipse", "Group", "Inset", "Line", "math", "MathML", "Polygon",
                   "Polyline", "Rectangle", "RoundRect", "TextFrame", "TextLine"],
    j, jMax, sName;

  if (obj == undefined) return false;
  jMax  = aGraphicObj.length;
  sName = obj.constructor.name
  for (j = 0; j < jMax; j++) {
    if (aGraphicObj [j] == sName ) { return true;}
  }
  return false;
} //--- end IsGraphicObject -----------------------------------------------------------------------

KLD_F.IsOneOf = function (sItems, sSearch) { // ==================================================
/*            Search a string (of some words) within another string, case dependent.
Arguments     sItems          String with items which probably contain sSearch
              sSearch         Item handled as full word
//            bWord           Consider sSearch as full word
Returns       Function value is number of occurrencies of sSearch
Called by     FindInSelection, GetCondFmtsList, WarnBackwardsSearch, ButtonFind, ButtonReplace,
              SelectedInCatalogB, SelectFindType
Calling       -
History       2021-07-09
*/
var nCount, eReg, aMatch;
  eReg = new RegExp ("\\b" + sSearch + "\\b", 'g'); // observe word boundaries
  aMatch = sItems.match (eReg, sSearch);          // replace by itself to get count
  nCount = aMatch == null ? 0 : aMatch.length;
  return nCount;
} //--- end IsOneOf -------------------------------------------------------------------------------

KLD_F.IsSelection = function (oDoc) { // ==========================================================
/*            Check whether a selection exists
Arguments     oDoc            Current document
Returns       True for existing selection/insertion point, else false
Called by     ButtonFind
Calling       -
Reference     Rick Quatro in https://community.adobe.com/t5/framemaker-discussions/why-is-a-text-selection-not-an-object/td-p/12921909#M74748
Used in       FMfindRepl
History       2022-05-10
*/
  if (oDoc.ObjectValid () === 1) {
    if (oDoc.TextSelection.beg.obj.ObjectValid () === 1) { return true;
      // alert ("There is an insertion point or text selected in the document.");
    } else if (oDoc.FirstSelectedGraphicInDoc.ObjectValid () === 1) { return true;
      // alert ("There is at least one graphic selected. (Aframe, …, TextFrame, TextLine, UnanchoredFrame");
    } else { return false;
      // alert ("Nothing is selected in the document.");
    }
  }
  return false;
} // --- end IsSelection --------------------------------------------------------------------------

KLD_F.MakerIniRead = function (sSection, sNames1) { //=============================================
/*            Read ini item(s) from maker.ini of the current FM session
Arguments     sSection        Section in maker.ini
//            sNames1…        Value name(s) as used in the ini-file
Returns       Function value is Array of read values; Not found items get value null
Called by     ButtonRegex
Calling       Message, GetIniItemIndex
Attention     If an item is found prior to the section line, it may be the wrong one.
              It is important to have unique item names in the files independently of sections
History       2020-12-29
*/
var appdata, jArg, nArguments, sIniFile, jItem, aLines=[], aMatch, bNotFound, oRegEx={},
    jSection, fSettings, sValueName, aValues=[];

  appdata   = app.UserSettingsDir;               // %appdata%\Adobe\FrameMaker\<version>
  sIniFile  = appdata + "\\maker.ini";
  bNotFound = true;
  oRegEx = {
    section: /^\s*\[\s*([^\]]*)\s*\]\s*$/,
    param: /^\s*([\w\.\-\_]+)\s*=\s*(.*?)\s*$/,
    comment: /^\s*;.*$/
  } ;
  fSettings = new File(sIniFile);
  fSettings.open("r");
  if (fSettings.error !== "") {
    KLD_F.Message ("E", KLD_F.UItxt.MakerIniRead.@msg_01, "MakerIniRead", fSettings);
    return "Ini-read-error";
  }
  nArguments = arguments.length;

  while(!fSettings.eof) {                         // read entire file into array ------------------
    aLines.push(fSettings.readln());
  }
  fSettings.close();

  bNotFound = true;
  nArguments = arguments.length;
  jSection = KLD_F.GetIniItemIndex (aLines, "["+sSection+"]");// Does section exist? -------------------
  if (jSection == null) {
    KLD_F.Message ("E", KLD_F.UItxt.MakerIniRead.@msg_02, "MakerIniRead", sSection, fSettings);
    return null;
  } else {                                        // we can look for the items
    for (jArg = 1; jArg < nArguments; jArg++) {   // 1, 2, 3 for n=4
      jItem = KLD_F.GetIniItemIndex (aLines, arguments[jArg]);
      if (jItem != null) {
        aMatch = aLines[jItem].match(oRegEx.param);
        sValueName = aMatch[1];
        aValues.push(aMatch[2]);
      } else {
        KLD_F.Message ("E", KLD_F.UItxt.MakerIniRead.@msg_03, "MakerIniRead", arguments[jArg], fSettings);
        aValues.push(null);
      }
    }
  }
  return aValues;
} // --- end MakerIniRead -------------------------------------------------------------------------

KLD_F.Message = function (msgType, msgTxt, parm1) { // ============================================
/*            Compose message in alert box or return as function value
Arguments     msgType  (1 character): I (info), any other character is interpreted for Error
                       for lower case characters the message is considered to bu use inline
                       (it is composed only and functions returns) - and has not title.
              msgText  String may conatin \n (newline), symbolic parameters %01 ... %09
                       \r (same as \n) and \t (tab) not working with Default layout behaviour
              msgTitle is derived from parm1, if message is not inline
Called by     Notify, BackupSettings, ConvertSelfClosingElements, GetXMLdata, MakerIniRead, OpenDocument,
              OpenHelpFile, OpenWebURL, SortLocal, WriteXMLdata, ApplyCharFmt, ApplyObjectStyle,
              ApplyPgfFmt, ApplyTblFmt, CheckForPotentialLoop, CollectUserStrings, ButtonReplace,
              ButtonSave, ButtonSaveNew, GetCatalogAny
Calling       -
Attention     If msgTxt is read in from XML (incorrectly) this may be of type XML.
              msgTitle Title for the window, will be prefixed wit application-name
              parm1 ... parm9:  strings replacing the symbolic parameters %01 � %09 (unused are skipped)
Returns       Composed message in case of lower case msgType, otherweise an alert message is issued
History       2020-11-04 test with VSCode and ESDebugger
*/
var bInline, diff, j, msgTitle, nArguments;

  if (typeof msgTxt == "xml" ) {msgTxt = msgTxt.toString();}
  if (msgTxt == undefined) {msgTxt = "--- empty / leer / vide ---";}
// Following two replacements only neccessary for read-in values, however they do not harm direct string parameters
  msgTxt = msgTxt.replace (/\\n/g , "\n");
  msgTxt = msgTxt.replace (/\\t/g , "\t");        // not working with ScriptUI Default layout behaviour
  if (msgType.match(/[a-z]/)) {                   // inline: parm1 is msgTitle
    bInline = true;
    diff    = 2;
  } else {
    msgTitle = arguments[2];
    bInline = false;
    diff    = 3;
  }
  nArguments = arguments.length - diff // inline has no title
  if (nArguments > 9) {
    nArguments = 9;                               // %10 would be interpreted as %01
  }
  for (j = 1; j <= nArguments; j++) {
    msgTxt = msgTxt.replace ("%0" + j, arguments[j+diff-1]);
  }
  if (bInline) {
    return msgTxt;
  }
  if (msgType.toUpperCase() == "I") {
    alert (msgTxt, msgTitle, false);
  } else {
    alert (msgTxt, msgTitle, true);
  }
} // --- end Message ------------------------------------------------------------------------------

KLD_F.ObjSelection = function (oXX) { // ==========================================================
/*            Define a TextRange from selected object
Argument      oXX             Object
Returns       Text Range
Called by     FindInSelection
Calling       -
Used in       FMfindRepl
History       2022-07-01
*/
var oTL, oTL2, oTR;

  oTL  = oXX.TextLoc;
  oTL2 = new TextLoc (oTL.obj, oTL.offset+1);
  oTR  = new TextRange (oTL, oTL2);
  return oTR;
} //--- end ObjSelection --------------------------------------------------------------------------

KLD_F.OpenDocument = function (filename, bVisible) { //============================================
/*            Open document or book�
Arguments     filename        fully qualified file name
              bVisible        false for hidden open
Returns       Function value is the document or book object
Called by     ButtonFind (for book components)
Calling       GetDocOpenProps, Message
Comment       Open errors handled according to GetDocOpenProps
History       2020-12-29
*/
var oDoc, openProps, openReturnParams;

  openProps = GetOpenDefaultParams();             // Get default property list for opening documents or books.
  openReturnParams = new PropVals();              // Get a property list to return any error messages.
  openProps = KLD_F.GetDocOpenProps (bVisible);         // false for invisible, ignore many errors
  oDoc = Open(filename, openProps, openReturnParams);// Attempt to open the document or book.
  if (oDoc.ObjectValid() === 1) {
    return oDoc;                                  // Return the document or book object.
  } else {                                        // Can't open, print the errors to the Console
    KLD_F.Message ("E", KLD_F.UItxt.OpenDocument.@msg_01, "OpenDocument", openReturnParams);
    return false;
  }
} //--- end OpenDocument --------------------------------------------------------------------------

KLD_F.OpenHelpFile = function () { // =============================================================
/*            Display the pdf with same name as this script ($.filename includes extension)
Called by     PalFindRepl
Calling       Message
Comment       No feedback in case the pdf does not exist => probably to be changed
              KLD_F.oHelpFile.numPages shoud display the number of pages in this pdf
*/
  if (KLD_F.oHelpFile.exists) {
    KLD_F.oHelpFile.execute();
  } else {
    KLD_F.Message ("I", KLD_F.UItxt.General.@NoHelpFile, KLD_F.APP_NAM, KLD_F.APP_NAM + ".pdf");
  }
} //--- end OpenHelpfFile -------------------------------------------------------------------------

KLD_F.OpenPdfAtDestination = function (pdfFile, location) { //=====================================
/*            POpen PDF at specified destination �
Arguments     pdfFile         full path in absoluteURI notation
              location        full definition - not just what was set in FM,
                              e.g. "nameddest=M8.newlink.FMcalcseries"
Returns       Function value is xxx
Called by     → currently not used
Calling       GetScriptPath
Comment       requires FMcalc_OpenPdf.exe in script directory
              (source: E:\_DDDprojects\FM-calc\test-scripts\OpenPdfWithParameters.ahk)
Reference     http://www.adobe.com/content/dam/acom/en/devnet/acrobat/pdfs/pdf_open_parameters.pdf
History       2020-12-29
*/
var path, tempFile;

  path = KLD_F.GetScriptPath () + "\\";
  tempFile = File (path + "FMxxx_tScript.bat");  // Create the calling batch file
  tempFile.open("w");
  tempFile.encoding = "UTF-8";
  tempFile.writeln ("REM Generated by FMcalc_Ancillary.jsx, function OpenPdfAtDestination");
  tempFile.writeln (path + "FMxxx_OpenPdf.exe " + '/A ' + location + ' "' + pdfFile.fsName + '"');
  tempFile.close();
  tempFile.execute ();
  tempFile.delete;
} //---  end OpenPdfAtDestination -----------------------------------------------------------------

KLD_F.OpenWebURL = function (url) {
/*            Open a webpage with the standard browser
Arguments     url             Complete URL, e.g. "http://www.regular-expressions.info/"
Called by     → may be used in PalRegEx
Calling       Message
Used in       FMfindRepl
History       2020-12-29
*/
	try{
		var URL = new File(Folder.temp + "/help.html");
		URL.open("w");
		URL.writeln('<html><HEAD><meta HTTP-EQUIV="REFRESH" content="0; url=' +  url + '"></HEAD></HTML>');
		URL.close();
		URL.execute();
	}catch(e){
		KLD_F.Message ("E", KLD_F.UItxt.OpenWebURL.@msg_01, "OpenWebURL", url);
	};
} //---  end OpenWebURL ---------------------------------------------------------------------------

KLD_F.PreserveSpace = function (oDoc, sReplace) { // ----------------------------------------------
/*            Preserve space for replacement in case of Smart Space ON
Arguments     oDoc            Current document
              sReplace        String for replacement
Returns       The string amended by a Space, if there one immediatly after the current selection.
Called by     ButtonReplace
Calling       GetText
History       2022-05-09
*/
var oTR = oDoc.TextSelection, oTL1, oTL2, oTR2, sWhat;

  oTL1 = new TextLoc (oTR.beg.obj, oTR.end.offset);
  oTL2 = new TextLoc (oTL1.obj, oTR.end.offset+1);
  oTR2 = new TextRange (oTL1, oTL2);
  sWhat = KLD_F.GetText(oDoc, oTR2);
  if (sWhat == " ") {
    if (oDoc.SmartSpaces) {sReplace += " ";}
  }
  oDoc.TextSelection = oTR;                       // Restore original selection
  return sReplace;
} //--- end PreserveSpace -------------------------------------------------------------------------

KLD_F.ReplaceInString = function (string, sSearch, sReplace, bCase, bWord, bRegEx) { // ===========
/*            Find substring with options and replae it
Arguments     string          String in which to search
              sSearch         String to be searched
              sReplace        String replacing the search
              bCase           true for case sentitive search
              bWord           true so search whole words only
              bRegEx          true to interprete sSearch as RegEx
Returns       The string with the replacement
Called by     ButtonReplace
Calling       -
History       2021-10-18
*/
var eReg0, eReg1, eReg2, eReg2, eReg3, eRegR, sResult, sFind;

  if (string === undefined) { string = ""; }
  if (!bRegEx) {
    sFind = sSearch.replace (/([\\\^\$\.\|\?\*\+\(\)\[\{])/g,"\\$1"); // escape the 12 syntactical chars
  }

  eReg0 = new RegExp (sFind, "i");                // case insesitive, no words
  eReg1 = new RegExp (sFind);                     // case sensitive, no words
  eReg2 = new RegExp (sFind + "\\b", "i");        // case insesitive, word
  eReg3 = new RegExp (sFind + "\\b");             // case sensitive, word

  if (bRegEx) {
    eRegR = new RegExp (sSearch);                 // interprete as RegEx
    sResult = string.replace (eRegR, sReplace);
  } else {
    if (!bCase && !bWord) {
      sResult = string.replace (eReg0, sReplace);
    } else if (bCase && !bWord) {
      sResult = string.replace (eReg1, sReplace);
    } else if (!bCase && bWord) {
      sResult = string.replace (eReg2, sReplace);
    } else {
      sResult = string.replace (eReg3, sReplace);
    }
  }
  return sResult;
} //--- end ReplaceInString -----------------------------------------------------------------------

KLD_F.Replacer1 = function (match, p1) { // =======================================================
/*            Helper function for replacement method
Arguments     match           Match found in XML file
              p1              First match group
Returns       Modified match
Called by     ConvertSelfClosingElements
Calling       -
Used in       FMfindRepl
History       2021-03-22
*/
  return "<" + p1 + "></" + p1 + ">";
} //--- end Replacer1 -----------------------------------------------------------------------------

KLD_F.RestoreLocation = function (oDoc, oLocation) { //============================================
/*            Restore currunt location to saved one
Arguments     oDoc            Current document
              oLocation       object created by SaveCurrentLocation
Returns       Function value is xxx
Called by     → currently not used
Calling       -
Attention     oLocation must not be a global variable - or a stack must be established
              Not tested what happens if this document is no more open at restore
Reference     Rick Quatro at https://forums.adobe.com/thread/2243878
Used in       FMbiblio, FMcalc, FMvars, FMgraph, FMnote …
History       2020-12-29
*/
var oPgf, tLoc, textRange;

  oDoc.CurrentPage = oLocation.page;
  oPgf = oLocation.pgf;
  if (oPgf.ObjectValid () === 1) {
    oDoc.TextSelection =   oLocation.txtRange;
  }
  oDoc.ScrollToText(textRange);
} //--- end RestoreLocation -----------------------------------------------------------------------

KLD_F.SaveCurrentLocation = function (oDoc) { //===================================================
/*            Save the currunt location for later restore
Arguments     oDoc            Current document
              asNames         Array of xy-names
Returns       Function value is Object to be used by RestoreLocation
Called by     → currently not used
Calling       -
Attention     Not tested what happens if this document is no more open at restore
Reference     Rick Quatro at https://forums.adobe.com/thread/2243878
Used in       FMbiblio, FMcalc, FMvars, FMgraph, FMnote …
History       xxx
*/
var oLocation = {};

  oLocation.pgf    = oDoc.TextSelection.beg.obj;
  oLocation.txtRange = oDoc.TextSelection;
  oLocation.page   = oDoc.CurrentPage;
  return oLocation;
} //--- end SaveCurrentLocation -------------------------------------------------------------------

KLD_F.SearchInString = function (string, iStart, sSearch, bCase, bWord, bRegEx) { // ==============
/*            Find substring with options
Arguments     string          String in which to search
              iStart          Index in string where to start the search
              sSearch         String to be searched
              bCase           true for case sentitive search
              bWord           true so search whole words only
              bRegEx          true to interprete sSearch as RegEx
Returns       The index of the found sSearch in string
Called by     FindInSelection, FindUserString
Calling       -
History       2021-10-18
*/
var eReg0, eReg1, eReg2, eReg2,  eReg3, eRegR, indx, sFind, sNext;

  if (string === undefined) { return; }
  sNext = string.substr (iStart);
  if (!bRegEx) {
    sFind = sSearch.replace (/([\\\^\$\.\|\?\*\+\(\)\[\{])/g,"\\$1"); // escape the 12 syntactical chars
  }

  eReg0 = new RegExp (sFind, "i");                // case insesitive, no words
  eReg1 = new RegExp (sFind);                     // case sensitive, no words
  eReg2 = new RegExp (sFind + "\\b", "i");        // case insesitive, word
  eReg3 = new RegExp (sFind + "\\b");             // case sensitive, word

  if (bRegEx) {
    eRegR = new RegExp (sSearch);                 // interprete as RegEx
    indx = sNext.search (eRegR);
  } else {
    if (!bCase && !bWord) {       indx = sNext.search (eReg0);
    } else if (bCase && !bWord) { indx = sNext.search (eReg1);
    } else if (!bCase && bWord) { indx = sNext.search (eReg2);
    } else {                      indx = sNext.search (eReg3);
    }
  }
  return indx + iStart;
} //--- end SearchInString ------------------------------------------------------------------------

KLD_F.SetDisplay = function (bWanted) { //=========================================================
/*            Set display to on or off
Arguments     bWanted         true => set display ON; false => display OFF
Returns       Function value is xxx
Called by     CollectUserStrings
Calling       -
Attention     Switching visibility to often lets this function fail.
              See https://forums.adobe.com/thread/2276988
Comment       xxx
Reference     Rick Quattro
Used in       FMfindRepl
History       FMbiblio, FMcalc, FMvars, ...
*/
  if (app.Displaying != bWanted) {
    app.Displaying = bWanted;
  }
} //--- end SetDisplay ----------------------------------------------------------------------------

KLD_F.SortNoCase = function (a, b) { // ===========================================================
/*            Compare function for case independent sorting
Arguments     a, b            Strings to be compared
Returns       Function value is 0 for equality, 1 for a > b; else -1
Called by     GetCharFmtsList, GetColourNameList, GetCondFmtsList, GetMarkerTypeList, GetPgfFmtsList,
              GetObjFmtList, GetTableFmtList, GetXrefFmtsList
Calling       -
History       2020-12-29
*/
// Arguments
// Called by
  if      (a.toLowerCase() >  b.toLowerCase()) {return 1;}
  else if (a.toLowerCase() == b.toLowerCase()) {return 0;}
  else                       {return -1;}
} //--- end SortNoCase ----------------------------------------------------------------------------

KLD_F.SortNoCase2 = function (a, b) { // ==========================================================
/*            Compare function for case independent sorting
Arguments     a, b            objects, the name of which to be compared
Returns       Function value is 0 for equality, 1 for a > b; else -1
Called by     GetVarNameList
Calling       -
History       2020-12-29
*/
// Arguments
// Called by
  if      (a.Name.toLowerCase() >  b.Name.toLowerCase()) {return 1;}
  else if (a.Name.toLowerCase() == b.Name.toLowerCase()) {return 0;}
  else                       {return -1;}
} //--- end SortNoCase ----------------------------------------------------------------------------

KLD_F.SortLocal = function (asNames, sLang) { // ==================================================
/*            Sort according to lexical order for the given language - partial solution only!
Arguments     asNames         Array of strings to be sorted
              sLang           ISO 2-char langugae code (de, en, fr)
Returns       Function returns sorted array. Original array is not touched.
Called by     GetSettingsList
Calling       CompareString, Message
Attention     This routine does not fulfill all requirements of a lexical order for any language.
              - No secondary sort level (req. for accented chars in French)
              - No 2-character equivalents for digraphs (Æ) or ä in German, ll = l in Spanish, ...
Comment       This is a rudimentary PolyFill for localeCompare with paramter and options (JS 5):
              referenceStr.localeCompare(compareString[, locales[, options]])
              e.g.: items.sort( (a, b) => a.localeCompare(b, 'fr', {ignorePunctuation: true}));
Reference     Klaus Göbel
              https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/localeCompare
              https://tc39.es/ecma402/#sup-String.prototype.localeCompare
              https://tc39.es/ecma402/#collator-objects
History       2020-11-23
*/

var j, n, swapped, aTemp = [], sSortDE, sSortEN, sSortFR, sSortOrder, temp;
// I have not found any reference for the complete sequences - even dictionaries do not list them
  sSortDE = "AaÄäÆæBbCcÇçDdEeÈèÉéÊêËëFfGgHhIiÌìÍíÎîÏïJjKkLlMmNnÑñOoÖöŒœÒòÓóÔôÕõPpQqRrSsßTtUuÜüÙùÚúÛûVvWwXxYyZzÿ";
  sSortEN = "AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZzÄÆäæÇçÈÉÊËèéêëÌÍÎÏìíîïÑñŒÒÓÔÖÕœòóôöõßÙÚÛÜùúûüÿ";
  sSortFR = "AaÄäÆæBbCcÇçDdEeÈèÉéÊêËëFfGgHhIiÌìÍíÎîÏïJjKkLlMmNnÑñOoÖöŒœÒòÓóÔôÕõPpQqRrSsßTtUuÜüÙùÚúÛûVvWwXxYyZzÿ";

  if ((asNames.constructor.name != "Array")) {
    KLD_F.Message ("E", KLD_F.UItxt.SortLocal.@msg_01, "SortLocal");
    return false;
  }
  if (sLang.toLowerCase() == "de") {
    sSortOrder = sSortDE;
  } else if (sLang.toLowerCase() == "en") {
    sSortOrder = sSortEN;
  } else if (sLang.toLowerCase() == "fr") {
    sSortOrder = sSortFR;
  } else {
    KLD_F.Message ("I", KLD_F.UItxt.SortLocal.@msg_02, "SortLocal");
    sSortOrder = sSortEN;
  }
  n = asNames.length;
  for (j = 0; j < n; j++) {
    aTemp[j] = asNames[j];
  }
  do {                                            // BubbleSort(sArray,sSortOrder); --- by Klaus Göbel
    swapped = false;
    for (j=0; j < n-1; j++) {
      if (KLD_F.CompareString(aTemp[j], aTemp[j+1], sSortOrder) == 1) {
        temp = aTemp[j];
        aTemp[j] = aTemp[j+1];
        aTemp[j+1] = temp;
        swapped = true;
      }
    }
  } while (swapped);
  return aTemp;

} //--- end SortLocal -----------------------------------------------------------------------------

KLD_F.StdSound = function () { // =================================================================
/*            Issue Sound for error messages
Calling       GetScriptPath, Helper application StdSound.exe (compiled AHK script), StdSound.wav
              The sound may be long, script continues only after the complete sound is played!
Attention     Helper app and sound filed must be in same directory
*/
var sPath, helperApp;

  sPath = KLD_F.GetScriptPath();
  helperApp = new File (sPath + "\\FMfindRepl\\StdSound.exe");
  if (helperApp.exists) {
    helperApp.execute();
  }
} // --- end StdSound -----------------------------------------------------------------------------

KLD_F.SwitchUIlanguageX = function (sLangFile) { // ===============================================
/*            Read the language relevant XML file to fill the global array of texts
Arguments     sLangFile       Base of the languaage xml file. will be appended
                              _de.xml etc.
                              For test pupose this may be the beginning of a real path name
Returns       Values in global XML object KLD_F.UItxt.
Called by     Main
Calling       GetXMLdata
Comment       I do not convert everything to strings, but use toString() where required
Used in       FMfindRepl
History       2020-11-16 test with VSCode and ESDebugger
*/
var appLang, sLangCode;

  appLang = app.Language;
//appLang = Constants.FV_LANG_FRENCH;             // set explictly for UI test
	switch (appLang) {
		case Constants.FV_LANG_ENGLISH:               // 0x01
		case Constants.FV_LANG_BRITISH: {             // 0x02
			KLD_F.UItxt = KLD_F.GetXMLdata (sLangFile + "_en.xml");
			sLangCode = "en";
			break;
		}
		case Constants.FV_LANG_GERMAN: {              // 0x03
			KLD_F.UItxt = KLD_F.GetXMLdata (sLangFile + "_de.xml");
			sLangCode = "de";
			break;
		}
		case Constants.FV_LANG_FRENCH: {              // 0x05
			KLD_F.UItxt = KLD_F.GetXMLdata (sLangFile + "_fr.xml");
			sLangCode = "fr";
			break;
		}
		default: {
			KLD_F.UItxt = KLD_F.GetXMLdata (sLangFile + "_en.xml");
			sLangCode = "en";
		}
	}
	return sLangCode;
} //--- end SwitchUIlanguageX ---------------------------------------------------------------------

KLD_F.UnformatClipboard = function (oDoc, oPgf) { // ==============================================
/*            Unformat the clip board to be able to preserve the target formats
Arguments     oDoc            Current document
              oPgf            Pgf into which the clipboard shall be pasted. This defines the
                              target format.
Returns       New clip board with formatting of target Pgf
Called by     ButtonReplace
Calling       ExecuteFCodes, ApplyPgfFmt
Attention     The clipboard must not contain multiple paragraphs.
              Only the last paragraph would be handled correctly.
Comment       There are no constants to be able to Paste without formatting or to apply the
              default paragraph format. There are also no properties for the clipboard -
              in particular the lenght of the Text Range in the clipboard can not be determined.
Used in       FMfindRepl
History       2022-05-28
*/
var oDoc, oPage, sPgfFmt, oLocation = {}, oPgfLocal, oTFr, oTL, oTR, MM = 185771;

  oDoc = KLD_F.oCurrentDoc;
  if (app.Displaying == true) {                   // Avoid flicker
    app.Displaying = false;
  }
  oLocation.pgf      = oDoc.TextSelection.beg.obj;// Save current location
  oLocation.txtRange = oDoc.TextSelection;
  oLocation.page     = oDoc.CurrentPage;
  sPgfFmt            = oPgf.Name;                 // Name of target paragraph format
// --- begin core function -----------------------
  oPage = oDoc.FirstMasterPageInDoc;              // Get first master page in document.
  oTFr = oDoc.NewTextFrame (oPage.PageFrame);     // Add a temporary text frame
  oTFr.Width  = 150 * MM;
  oTFr.Height =  50 * MM;
  oPgfLocal = oTFr.FirstPgf;
  oTR = new TextRange;
  oTR.end.obj = oTR.beg.obj = oPgfLocal;
  oTR.end.offset = oTR.beg.offset = 0;
  oDoc.TextSelection = (oPgfLocal,oTR);
  oDoc.Paste(0);                                  // Paste the clip board
  oTR.beg.obj = oPgfLocal;                        // ClipBoard may contain more than 1 ¶
  oTR.beg.offset = 0;
  oTR.end.obj = oPgfLocal;
  oTR.end.offset = Constants.FV_OBJ_END_OFFSET;
  oDoc.TextSelection = oTR;
  KLD_F.ExecuteFCodes (["SelectAll","CharacterDefaultPgfFont"]); // remove the overwrites¶
  KLD_F.ApplyPgfFmt (oDoc, oPgfLocal, sPgfFmt);;        // Complete format must be applied
  oDoc.Copy (0);                                  // Copy selection to clipboard not interactive
  oTFr.Delete ();                                 // Delete the helper text frame
// --- end core function -------------------------
  oDoc.CurrentPage = oLocation.page;              // Restore current location
  oPgf = oLocation.pgf;
  if (oPgf.ObjectValid () === 1) {
    oDoc.TextSelection =   oLocation.txtRange;
  }
  oDoc.ScrollToText(oLocation.txtRange);
  if (app.Displaying == false) {                  // Restore previous state
    app.Displaying = true;
    app.ActiveDoc.Redisplay();
  }
} // --- end UnformatClipBoard --------------------------------------------------------------------

KLD_F.UnGroup = function (oGroup) { // ============================================================
/*            Dissolve the group
Arguments     oGroup          Group object to be dissolved
Returns       number of ungrouped objects
Called by     ButtonFind
Calling       -
History       2021-07-08
*/
var aItems = [], j, k= 0, oGraphic;

  oGraphic = oGroup.FirstGraphicInGroup;
  while (oGraphic.ObjectValid()) {
    aItems[k] = oGraphic;
    oGraphic = oGraphic.NextGraphicInGroup;
    k ++;
  }

  for (j = 0; j < k; j++) {
    aItems[j].GroupParent = 0;
  }
  return k;
} //--- end UnGroup -------------------------------------------------------------------------------

KLD_F.VerifyRegEx = function (sRegEx) { // ========================================================
/*            Verify the regular expression
Arguments     -
Returns       Function value is true or false
Called by     PalRegEx: ButtonVerify
Calling       -
Reference     C# Cookbook by O'Reilly Media, Inc. 2004. ISBN: 9780596003395
              Section 8.3: Verifying the Syntax of a Regular Expression.
Used in       FMfindRepl
History       2021-03-09
*/
var aItems, isValid = true, sValue = "";
  if ((sRegEx != null) && (sRegEx.length > 0)) {
    try {
      aItems = sValue.match (sRegEx);
    }
    catch (ArgumentException) { // BAD PATTERN: Syntax error
      isValid = false;
    }
  } else {                      //BAD PATTERN: Pattern is null or blank
    isValid = false;
  }
  return (isValid);
} // --- end VerifyRegEx --------------------------------------------------------------------------

KLD_F.WriteXMLdata = function (xmlData, sXmlFile) { // ============================================
/*            Write XML data to file
Argements     xmlData         the full structure to be written.
              sXmlFile        fully qualified path of XML file or just the file name
                              If no filepath is given (no slash or backslash present),
                              then the file is assumed in the same directory as the script.
                              The file is created anew in UTF-8 with BOM
Called by     ButtonDelete, ButtonSave, ButtonSaveNew
Calling       Message
Reference     https://www.davidebarranca.com/2013/03/presets-management-with-dropdownlist-and-xml-in-extendscript/
              Klaus Göbel in https://community.adobe.com/t5/framemaker/xml-and-unicode-a-quarrelling-couple/td-p/11677704
History       2020-12-14
*/
var BOM = "\uFEFF", fXmlFile;

  if ((sXmlFile == null) || (sXmlFile == undefined)) {
    return false;
  } else if (sXmlFile.indexOf("/") > 0 || sXmlFile.indexOf("\\") > 0) { // fully qualified path
    fXmlFile = new File(sXmlFile);
  } else {
    fXmlFile = new File($.fileName.replace (/[^\\\/]+$/i , sXmlFile));  // file in script folder
  }
  try {
    fXmlFile.open('w');
    fXmlFile.encoding = "UTF-8";
    fXmlFile.write(BOM);
    fXmlFile.write(xmlData);
    fXmlFile.close();
  } catch (e) {
    KLD_F.Message ("E", KLD_F.UItxt.WriteXMLdata.@msg_01, "WriteXMLdata", sXmlFile, e.message);
    return false;
  }
  return true;
} // --- End WriteXMLdata -------------------------------------------------------------------------
