﻿/* --- FMfindRepl_DocObj.jsx ------ UTF-8 ------------------------------------------------------
              Functions related to the FM object model
Comment       to be included in FMfindRepl.jsx
History       2020-12-29
*/

KLD_F.AdvanceTextLoc = function (oParent) { // ====================================================
/*            Advance the text location after the anchor to be able to continue a search
Arguments     oParent         Parent object (AFrame, Tbl) which has an anchor
Returns       oDoc.TextSelection
Called by     RemoveOverride
Calling       -
Used in       FMfindRepl
History       2022-06-18
*/
var oDoc = KLD_F.oCurrentDoc, oTL, oTR;

  if (oParent != undefined ) {
    oTL = oParent.TextLoc;
    oTL.offset += 1;
    oTR = new TextRange(oTL, oTL);
    oDoc.TextSelection = oTR;
  }
} // --- end AdvanceTextLoc -----------------------------------------------------------------------

KLD_F.ApplyCharFmt = function (oDoc, oTR, sFormat) { // ===========================================
/*            Apply character format to text range
Arguments     oDoc            Document to be handled, not necessarily the current document
              oTR             Text range to receive the format
              sFormat         Name of character format
Returns       -
Called by     ButtonReplace
Calling       Message
Reference     Rick Quatro
Used in       FMnotes etc.
History       2016-02-03
              2022-06-16 check for catalogged name
*/
var oCharFmt;

  oCharFmt = oDoc.FirstCharFmtInDoc;              // is it in catalogue?
  while (true) {
    if (oCharFmt.Name == sFormat) { break;}       // it's in the catalogue
    oCharFmt = oCharFmt.NextCharFmtInDoc;
    if (!oCharFmt.ObjectValid()) {                // it's not cataloggued
      KLD_F.Message ("E", KLD_F.UItxt.ApplyCharFmt.@msg_01, "ApplyCharFmt", sFormat);
      return false;
    }
  }

  oCharFmt = oDoc.GetNamedCharFmt(sFormat);
  oProps = oCharFmt.GetProps();
  oDoc.SetTextProps (oTR, oCharFmt.GetProps());
  return true;
} // --- end ApplyCharFmt -------------------------------------------------------------------------

KLD_F.ApplyCondFmtToRow = function (oDoc, oRow, sCondFmt) { // =====================================
/*            Apply condition format to text range
Arguments     oDoc            Document to be handled
              oRow            Table row to receive the format
              sCondFmt        Name of Conditon format (was selected from catalogue)
Returns       -
Called by     ButtonReplace
Calling       -
Reference     Russ Ward in https://community.adobe.com/t5/framemaker-discussions/applying-conditional-text-with-setprops/m-p/6122283
              http://frameautomation.com/category/extendscript/page/3/
Used in       FMfindRepl
History       2022-08-02
*/
var aoCondFmts, j, n, oCondFmt, oProps, iPropIndex;

  oCondFmt = oDoc.GetNamedCondFmt(sCondFmt);
  if(!oCondFmt.ObjectValid()) return;             // emergency exit

  aoCondFmts = oRow.InCond;                       // Current list of applied conditions
  n = aoCondFmts.length;
  for (j=0; j< n; j++) {
    if (aoCondFmts[j].Name == sCondFmt) {
      return;                                     // format exists already
    }
  }
  aoCondFmts.push (oCondFmt);                     // add the new one
  oRow.InCond = aoCondFmts;                       // Conditions were added; apply the updated list
} // --- end ApplyCondFmtToRow --------------------------------------------------------------------

KLD_F.ApplyCondFmtToText = function (oDoc, oTR, sCondFmt) { // =====================================
/*            Apply condition format to text range
Arguments     oDoc            Document to be handled
              oTR             Text range to receive the format
              sCondFmt         Name of Conditon format (was selected from catalogue)
Returns       -
Called by     ButtonReplace
Calling       -
Reference     Russ Ward in https://community.adobe.com/t5/framemaker-discussions/applying-conditional-text-with-setprops/m-p/6122283
              http://frameautomation.com/category/extendscript/page/3/
              FDK Programmers Guide "Applying conditions to text"
Used in       FMfindRepl
History       2022-08-02
*/
var iCondID, j, n, oCondFmt, oProps, iPropIndex;

  oCondFmt = oDoc.GetNamedCondFmt(sCondFmt);
  if(!oCondFmt.ObjectValid()) return;             // emergency exit
  iCondID  = oCondFmt.id;

  oProps = oDoc.GetTextPropVal (oTR.beg, Constants.FP_InCond);
  n = oProps.propVal.osval.length;
  for (j=0; j< n; j++) {
    if (oProps.propVal.osval[j].Name == sCondFmt) {
      return;                                     // format exists already
    }
  }
  oProps.propVal.osval.push(oCondFmt);
  oProps.propVal.isval.push(iCondID);
  oDoc.SetTextPropVal (oTR, oProps);              // ???
} // --- end ApplyCondFmtToText -------------------------------------------------------------------

KLD_F.ApplyObjectStyle = function (oDoc, oObject, sStyle) { // ====================================
/*            Purpose_of_this_woderful_function…
Arguments     oDoc            Document to be handled, not necessarily the current document
              oObject         Object to receive the style
              sStyle          Name of object style
Returns       false in case of error; else true
Called by     RemoveOverride, ButtonReplace
Calling       Message
Reference     https://forums.adobe.com/thread/1711161
Used in       FMfindRepl
History       2022-06-16 check for catalogged name
*/
var oObjStyle, oGraphicsFmt, oProps, sStyle;

  oGraphicsFmt = oDoc.FirstGraphicsFmtInDoc;      // is it in catalogue?
  while (true) {
    if (oGraphicsFmt.StyleTag == sStyle) { break;}// it's in the catalogue
    oGraphicsFmt = oGraphicsFmt.NextGraphicsFmtInDoc;
    if (!oGraphicsFmt.ObjectValid()) {            // it's not cataloggued
      KLD_F.Message ("E", KLD_F.UItxt.ApplyObjectStyle.@msg_01, "ApplyObjectStyle", sStyle);
      return false;
    }
  }
  oObjStyle = oDoc.GetNamedGraphicsFmt (oGraphicsFmt.StyleTag);
  oProps = oObjStyle.GetProps();
  oObject.SetProps(oProps);
  return true;
} // --- end ApplyObjectStyle ---------------------------------------------------------------------

KLD_F.ApplyPgfFmt = function (oDoc, oPgf, sFormat) { // ===========================================
/*            Apply paragraph format to paragraph
Arguments     oDoc            Document to be handled, not necessarily the current document
              oPgf            Paragraph to receive the format
              sFormat         Name of paragraph format
Returns       false in case of error; else true
Called by     UnformatClipboard, RemoveOverride, ButtonReplace
Calling       Message
Reference     https://forums.adobe.com/thread/1711161
Used in       FMfindRepl, FMnotes etc.
History       2019-09-13
              2022-06-16 check for catalogged name
*/
var msg_01, oPgfFmt, oProps;

  oPgfFmt = oDoc.FirstPgfFmtInDoc;                // is it in catalogue?
  while (true) {
    if (oPgfFmt.Name == sFormat) { break;}        // it's in the catalogue
    oPgfFmt = oPgfFmt.NextPgfFmtInDoc;
    if (!oPgfFmt.ObjectValid()) {                 // it's not cataloggued
      KLD_F.Message ("E", KLD_F.UItxt.ApplyPgfFmt.@msg_01, "ApplyPgfFmt", sFormat);
      return false;
    }
  }
  oPgfFmt = oDoc.GetNamedPgfFmt(sFormat);
  oProps = oPgfFmt.GetProps();
  oPgf.SetProps(oProps);
  return true;
} //--- end ApplyPgfFmt ---------------------------------------------------------------------------

KLD_F.ApplyTblFmt = function (oDoc, oTbl, sFormat) { // ===========================================
/*            Apply table format to table
Arguments     oDoc            Document to be handled, not necessarily the current document
              oTbl            Table to receive the format
              sFormat         Name of table format
Returns       false in case of error; else true
Called by     RemoveOverride, ButtonReplace
Calling       Message
Used in       FMfindRepl
History       2022-06-16 check for catalogged name
*/
var msg_01, oTblFmt, oProps;

  oTblFmt = oDoc.FirstTblFmtInDoc;                // is it in catalogue?
  while (true) {
    if (oTblFmt.TblTag == sFormat) { break;}      // it's in the catalogue
    oTblFmt = oTblFmt.NextTblFmtInDoc;
    if (!oTblFmt.ObjectValid()) {                 // it's not cataloggued
      KLD_F.Message ("E", KLD_F.UItxt.ApplyTblFmt.@msg_01, "ApplyTblFmt", sFormat);
      return false;
    }
  }
  oTblFmt = oDoc.GetNamedTblFmt(sFormat);
  oProps = oTblFmt.GetProps();
  oTbl.SetProps(oProps);
  return true;
} //--- end ApplyTblFmt ---------------------------------------------------------------------------

KLD_F.BtnFindByEnter = function (oEvt) { // =======================================================
/*            Activate BtnFind when enter is used in Find field
Arguments     oEvt            Event object to be handled.
Globals       KLD_F.evBuffer  Takes the sequence of typed keys, Backspace removes last entered
              evParm          Must be defined in the target (The element object where the event occurres)
                              This defines the array of items in the list.
Returns       ???
Called by     wPalF.g02.eFindWhat in PalFindRepl
Reference     function ListSelector
Used in       FMfindRepl
History       2023-07-27
*/
var kName;
  kName = oEvt.keyName.toLowerCase();
  if (kName == "enter"|| kName == "tab") {        // activate the Find button
    wPalF.g08.btnFind.active   = true;
  }
} // --- end BtnFindByEnter -----------------------------------------------------------------------

KLD_F.CheckForPotentialLoop = function (sSearch, sReplace, bWord) { // ============================
/*            Check for potential loop situation for find/replacce
Arguments     sSearch           Set of properties as defined in the current object
              sReplace          Set of properties as defined in the catalobue
              bWord             True if searching with word constraint
Return        True for OK button presses, False for No (default)
Called by     ButtonReplFnd, ButtonReplAll
Calling       Message, DisplayInfoF
Comment       This is not specific to FMfindRepl. It may also happen to the standard FM process
              See documentation FMfindRepl
Reference     Own experiments
Used in       FMfindRepl
History       2022-07-15
*/
var bOK = true, msg, z;

  if (!bWord) {
    z = sReplace.indexOf (sSearch, 0);
    if (sReplace.indexOf (sSearch, 0) >= 0) {
      msg = KLD_F.Message("e", KLD_F.UItxt.CheckForPotentialLoop.@msg_02, sSearch, sReplace);
      KLD_F.DisplayInfoF (msg, "red");
      bOK = confirm (msg, false, KLD_F.UItxt.CheckForPotentialLoop.@msg_01.toString());
    }
    return bOK;
  }
  return true;
} //--- end CheckForPotentialLoop -----------------------------------------------------------------

KLD_F.CollectUserStrings = function (oDoc, aoUserStrings) { // ====================================
/*            Collect user strings from certain object types
Arguments     oDoc            Current document
              aoUserStrings   Array into which to collect the objects
Global Var    aUserStrTypes   Array of names of objects from whic User Strings are searched.
                              ["AFrame", "Tbl"];
Returns       aoUserStrings   [object1, object2 ...]
                              object.ObjType  type of object
                              object.String   user string of object
                              object.TL       text location of object anchor
Called by     SelectFindType
Calling       UserString, SetDisplay, GetFindParameters, GetPage, Message
Comment       User strings are defined for all of the following object types:
              AFrame, app, XRef, Var, TblFmt, Tbl, SubCol, RulingFmt, Row, PfgFmt, Pgf, Marker,
              Fn, Flow, Element, Doc, CondFmt, ConbinedFontDefn, Color, CharFmt, Cell, BookComponent
              Therefore it is not senseful to check here all of these types.
              Only the doc, flows, anchores frames and tables are searched.
Reference     The first-next method does not work for anchored frames. I had to resort to the find method.
Used in       FMfindRepl
History       2022-06-25
*/
var j, k, n, sFound, asTypes = KLD_F.aUserStrTypes,
    oObject, oFindParms, oFrameTI, oPage, oPgf, oTextItem, oTL, oTL2, oTRfound, oUstring = {};

  aoUserStrings.length = 0;                       // clear array

  sFound = oDoc.UserString;                       // Doc is always considered
  oPgf = oDoc.MainFlowInDoc.FirstTextFrameInFlow.FirstPgf;
  oTL  = new TextLoc (oPgf, 0);
  oUstring = new KLD_F.UserString("Doc", oDoc, sFound, oTL);
  aoUserStrings.push (oUstring);

  KLD_F.SetDisplay (false);
//KLD_F.DisplayInfoF (KLD_F.Message("i", KLD_F.UItxt.CollectUserStrings.@msg_03), "black");
  for (j = 0; j < asTypes.length; j++) {
    if        (asTypes[j] == "AFrame") {          // Collect Anchored Frames (Body pages only) ----
      oObject = oDoc.MainFlowInDoc.FirstTextFrameInFlow.FirstPgf; // body pages only
      oTL = new TextLoc (oObject, 0);
      oDoc.TextSelection = new TextRange (oTL, oTL);;
      oFindParms = KLD_F.GetFindParameters ("ANCFRM", "", 0, false, false, false)
  //                                       (sType, sSearch, iMode, bWord, bCase, bBack)
      oTRfound = oDoc.Find(oTL, oFindParms);
      while (oTRfound.beg.obj.ObjectValid()) {
        oFrameTI = oDoc.GetTextForRange (oTRfound, Constants.FTI_FrameAnchor);
        for (k = 0; k < oFrameTI.length; k++) {
          oTextItem = oFrameTI[k];
          oObject = oTextItem.obj;
          sFound = oObject.UserString;
          oTL  = oObject.TextLoc;
          oUstring = new KLD_F.UserString("AFrame", oObject, sFound, oTL);
          aoUserStrings.push (oUstring);
        }
        oTRfound = oDoc.Find(oTL, oFindParms);
      }
    } else if (asTypes[j] == "Tbl") {             // Collect Tables (Body pages only) --------------
      oObject = oDoc.FirstTblInDoc;
      while (oObject.ObjectValid()) {
        sFound = oObject.UserString;
        oTL  = oObject.TextLoc;
        oPage = KLD_F.GetPage (oDoc, oObject);
        if (oPage.constructor.name == "BodyPage") {
          oUstring = new KLD_F.UserString("Tbl", oObject, sFound, oTL);
          aoUserStrings.push (oUstring);
        }
        oObject = oObject.NextTblInDoc;
      }
    } else {
      KLD_F.Message ("E", KLD_F.UItxt.CollectUserStrings.@msg_02, "CollectUserStrings", asTypes[j]);
    }
  }
  KLD_F.SetDisplay (true);
  if (aoUserStrings.length < 2) {
      KLD_F.Message ("E", KLD_F.UItxt.CollectUserStrings.@msg_01, "CollectUserStrings", asTypes);
  }
  n = aoUserStrings.length;
  KLD_F.nUserStrings = n;                         // set global variable
  KLD_F.jCurrentUStr = 0;                         // for first search
  return n;
} // --- end CollectUserStrings -------------------------------------------------------------------

KLD_F.CompareColour = function(oColour, kColour, sColour) { // ====================================
/*            compare colours
Arguments     kReserved       0  for user colour
                                   >0 nr of reserved colour
              sColour         Name of colour to be found (user colour only)
Returns       true if colour found, else false
Called by     FindColourUse
Calling       -
Used in       FMfindReplace
History       2021-02-19
*/
  if (kColour == 0) {
    if (!sColour) { //  == undefined
      if (!oColour.ObjectValid()) {
        return true;                              // uncoloured object found
      }
    } else {
      if (oColour.ReservedColor == 0 && oColour.Name == sColour) {return true;} // user coulour found
    }
  } else {
    if (oColour.ReservedColor == kColour) {
      return true;                                // system colour found
    }
  }
  return false;
} // --- end CompareColour ------------------------------------------------------------------------

KLD_F.FindColourUse = function (kReserved, sColourIn) { // ========================================
/*            Purpose_of_this_woderful_function
Arguments     kReserved       0  for user colour
                                   >0 nr of reserved colour
              sColour         Name of colour to be found (user colour only)
Returns       true if colour found, else false
Called by     ButtonFind
Calling       IsGraphicObject, GetNextObjectInDoc, CompareColour, DisplayObject, DisplayInfoF
Used in       FMfindReplace
History       2021-02-19
*/
var oDoc, obj, sColour = sColourIn;

  KLD_F.iError = 0;
  oDoc = KLD_F.oCurrentDoc;
  obj = KLD_F.oLstObjFound.object;                // start object
  if (obj == undefined) return undefined;

  if (sColourIn == null || sColourIn == undefined) {sColour = undefined;}
  if (KLD_F.IsGraphicObject(obj)) {               // Graphic object -------------------------------
    while (obj != undefined) {
      if (obj.constructor.name == "UnanchoredFrame") { // far to many of thsse uncoloured
        obj = KLD_F.GetNextObjectInDoc(obj);
        continue;
      }
      if (KLD_F.CompareColour(obj.Color, kReserved, sColour)) {
        break;
      } else {
        obj = KLD_F.GetNextObjectInDoc(obj);
      }
    }
    if (obj != undefined) {
      KLD_F.DisplayObject (obj);
      KLD_F.oLstObjFound.object  = obj;
      return obj;
    } else {                                      // try next type of object
      obj = oDoc.FirstTblInDoc.FirstRowInTbl.FirstCellInRow;
    }
  }

  if (obj.constructor.name == "Cell") {           // Table cell -----------------------------------
    while (obj != undefined) {
      if (KLD_F.CompareColour(obj.CellOverrideShading, kReserved, sColour) && obj.CellIsStraddled == 0) {
        break;
      } else {
        obj = KLD_F.GetNextObjectInDoc (obj);     // may go to next table
      }
    }
    if (obj != undefined) {
      KLD_F.DisplayObject (obj);
      KLD_F.oLstObjFound.object  = obj;
      return obj;
    } else {                                      // try next type of object
      obj = oDoc.FirstPgfInDoc;
    }
  }
  if (obj.constructor.name == "Pgf")  {           // Paragraph ------------------------------------
    while (obj != undefined) {
      if (KLD_F.CompareColour(obj.Color, kReserved, sColour) ||
        KLD_F.CompareColour(obj.BkColor, kReserved, sColour) ||
        KLD_F.CompareColour(obj.PgfBoxColor, kReserved, sColour)) {
        if(obj.InTextObj.constructor.name == "Cell" && obj.InTextObj.CellIsStraddled > 0) {
          KLD_F.DisplayInfoF (KLD_F.UItxt.FindColourUse.@msg_01, "red");
        }
        break;
      } else {
        obj = KLD_F.GetNextObjectInDoc(obj);
      }
    }
    if (obj != undefined) {
      KLD_F.DisplayObject (obj);
      KLD_F.oLstObjFound.object  = obj;
      return obj;
    }
  }
    return undefined;
} //--- end FindColourUse -------------------------------------------------------------------------

KLD_F.FindEmptyTextLine = function (oLastObj) { // ================================================
/*            Find and select empty text line
Arguments     oLastObj        Last found object to continue search
Returns       Function value is object:
              object.obj      the found graphic object or undefined for "nothing found"
              object.grouped  false | true
Called by     ButtonFind
Calling       GetText, IsEmptyString
Comment       Grouped text lines can be found, indication is just by the blinking text cursor.
              To delete them they must be undgrouped first.
              Ungrouped text lines are indicated by their 8 handles.
Reference     DeleteEmptyTextlines.fsl by Michael Müller-Hillebrand (200-04-22)
Used in       FMfindRepl
History       2021-02-09
              2022-11-15 On active selection start afresh
*/
var bFound = false, oDoc, oGraphic, oTL, sText, oTR;

  oDoc = KLD_F.oCurrentDoc;
  if (KLD_F.IsSelection(oDoc)) {                  // 2022-11-15
    oGraphic = oDoc.FirstGraphicInDoc;
  } else {
    oGraphic = oLastObj.object;
  }
  KLD_F.iError = 0;

  while (oGraphic.ObjectValid()) {
    if (oGraphic.constructor.name == "TextLine") {
      sText = KLD_F.GetText(oDoc, oGraphic);
      if (KLD_F.IsEmptyString(sText)) {           // string is empty or blank
        bFound = true;
        if (oGraphic.GroupParent.ObjectValid()) {
          KLD_F.oLstObjFound.grouped = true;
        } else {
          KLD_F.oLstObjFound.grouped = false;
        }
//      KLD_F.DisplayObject(oGraphic);            // Does not indicate grouped Textlines
        oTL = new TextLoc (oGraphic, 0);
        oTR = new TextRange(oTL, oTL);
        oDoc.ScrollToText(oTR);
        oGraphic.GraphicIsSelected = true;        // indicate it
//      oDoc.Redisplay();                         // no help against ghostig

        break;
      }
    }
    oGraphic = oGraphic.NextGraphicInDoc;
  }
  KLD_F.oLstObjFound.object  = oGraphic;
  if (bFound) {
    return KLD_F.oLstObjFound;
  } else {
    return undefined;                             // nothing found
  }
} //--- end FindEmptyTextLine ---------------------------------------------------------------------

KLD_F.FindInSelection = function (sFType, sSearch, iMode, bWord, bCase, bBack) { // ===============
/*            Find the first relevant object in the current text selection
Argument      sFType          Textual ID of Find Type. See Table asFindType
              sSearch         what to search for - used only for certain types
              iMode           0 Simple search
                              1 Wildcard search; excludes RegEx and simple Search
                              2 Regex search; excludes bWord and bBack; RegexFlavour is set to Perl
              bWord           true to constrain search to whole word
              bCase           true to consider case in search
              bBack           Not used
Returns       Text Range
Called by     ButtonFind
Calling       IsOneOf, FindSomething, DisplayInfoF, SearchInString, ObjSelection, GetText
Comment       Searched text is not found in embedded objects: AFrame *, table, cross reference *, variable *
              * same in FM-F/C
Attention     GetText eliminates automatic hyphens!
Reference     RQ in https://community.adobe.com/t5/framemaker-discussions/find-something-in-a-text-selection/td-p/13040904
Used in       FMfindRepl
History       2022-07-01
*/
var aTbls = [], aTfound, aPgfs = [],
    bRegEx = false, j, jOffset, oDoc = KLD_F.oCurrentDoc,
    oPgf, oTL, oTL2, aTi, oTR = oDoc.TextSelection, oTRold, oTRfound, oXX,
    regEx, sText = "";

  if (iMode == 2) {bRegEx = true; }

  if (KLD_F.IsOneOf ("AUTHYP CHRTAG CONTAG OSTFMO OSTTAG PGFFMO RUBANY ", sFType) > 0) {
/* For these Find Types no FTI constants for GetForRange are defined → Divide and conquer:
  Separate the given Text Range into paragraphs forming distinct Text Ranges.
  A general selection may provide incomplete first and last paragraphs, so the algorithm should
  store Text Ranges for the Search - but these could not be identified by unique.
  Hence I work with complete paragraphs. → Defining aTi with constants (FTI_PgfBegin | FTI_PgfEnd).
  This will create duplicate entries for the complete paragraphs → room for optimisation.
*/
    oTRold = oTR;
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_PgfBegin | Constants.FTI_PgfEnd);
    for (j = 0; j < aTi.length; j++) {
      oPgf = aTi[j].obj;
      aPgfs.push (oPgf);
    }
    for (j = 0; j < aPgfs.length; j++) {
      oPgf = aPgfs[j];
      oTL  = new TextLoc (oPgf, 0);
      oDoc.TextSelection = new TextRange (oTL, oTL); // req. in FindSomething
      oTR = KLD_F.FindSomething (sFType, sSearch, 0, false, false, false); // takes TextSelection
      if (oTR.beg.obj.Unique == oPgf.Unique) {
        return oTR;
      }
    }
    KLD_F.DisplayInfoF (KLD_F.UItxt.ButtonFind.@msg_07.toString(), "red"); // not found (anymore)
    oDoc.ScrollToText (oTRold);
    return oTRold;

  } else if (sFType == "ANCFRM") { // ----------- Anchored Frames --------------------------------- OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_FrameAnchor);

  } else if (sFType == "CHRFMO") { // ----------- Character Formats ------------------------------- OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_CharPropsChange | Constants.FTI_PgfBegin);
    if (aTi != undefined) {
      oXX = aTi[0].obj;                           // invalid objects in [1] and [2]
      oTL  = new TextLoc (oXX, aTi[1].offset);
      oTL2 = new TextLoc (oXX, aTi[2].offset);
      oTR  = new TextRange (oTL, oTL2);
      return oTR;
     }
    return undefined;

  } else if (sFType == "FOONOT") { // ----------- Footnote --------------------------------------- OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_FnAnchor);

  } else if (sFType == "MRKANY") { // ----------- Markers ----------------------------------------- OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_MarkerAnchor);
  } else if (sFType == "MRKTXT") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_MarkerAnchor);
    for (j = 0; j < aTi.length; j++) {
      oXX = aTi[j].obj;
      if (KLD_F.SearchInString (oXX.MarkerText, 0, sSearch, bCase, bWord, bRegEx) < 0) {
        continue;
      } else {
        return KLD_F.ObjSelection (oXX);
      }
    }
    return undefined;
  } else if (sFType == "MRKTYP") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_MarkerAnchor);
    for (j = 0; j < aTi.length; j++) {
      oXX = aTi[j].obj;
      if (oXX.MarkerTypeId.Name != sSearch) {
        continue;
      } else {
        return KLD_F.ObjSelection (oXX);
      }
    }
    return undefined;

  } else if (sFType == "PGFTAG") { // ----------- Paragraph tags ---------------------------------- OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_PgfBegin);
    for (j = 0; j < aTi.length; j++) {
      oXX = aTi[j].obj;
      if (oXX.Name != sSearch) {
        continue;
      } else {
        oTL  = new TextLoc (oXX, 0);
        oTL2 = new TextLoc (oXX, Constants.FV_OBJ_END_OFFSET);
        return new TextRange (oTL, oTL2);
      }
    }
    return undefined;

  } else if (sFType == "TBLANY") { // ----------- Tables ------------------------------------------ OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_TblAnchor);
  } else if (sFType == "TBLFMO") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_TblAnchor);
    for (j = 0; j < aTi.length; j++) {
      oXX = aTi[j].obj;
      aTbls.push (KLD_F.ObjSelection (oXX)); // TR's
    }
    for (j = 0; j < aTbls.length; j++) {
      oTR = aTbls[j];
      oDoc.TextSelection = oTR;
      oTR = KLD_F.FindSomething ("TBLFMO", "", 0, false, false, false); // takes TextSelection
      if (oTR != undefined) {return oTR;}
    }
    return undefined;
  } else if (sFType == "TBLTAG") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_TblAnchor);
    for (j = 0; j < aTi.length; j++) {
      oXX = aTi[j].obj;
      if (oXX.TblTag != sSearch) {
        continue;
      } else {
        return KLD_F.ObjSelection (oXX);
      }
    }
    return undefined;

  } else if (sFType == "TINANY") { // ----------- Insets ------------------------------------------ OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_TextInsetBegin);
    oXX= aTi[0].obj;
    oTRfound = oXX.TextRange;
    return oTRfound;
  } else if (sFType == "TINUNR") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_TextInsetBegin);
    for (j = 0; j < aTi.length; j++) {
      if (aTi[j].sData == undefined) {
        return aTi[j].obj.TextRange;
      } else {continue;}
    }
    return undefined;

  } else if (sFType == "TXTTXT") { // ----------- Text -------------------------------------------- OK
// Searched text is not found in embedded objects: AFrame *, table, cross reference *, variable *
// * same in FM-F/C
    regEx = new RegExp(sSearch);	                // create RegEx from string
    sText = KLD_F.GetText (oDoc, oTR);            // Constants.FTI_String
    aTfound = sText.match(regEx);
    if (aTfound != null) {
      oTRfound = new TextRange();
      oTRfound.beg.obj = oTR.beg.obj;
      oTRfound.beg.offset = oTR.beg.offset + aTfound.index - aTfound[0].length;
      if (oTRfound.beg.offset < 0) {oTRfound.beg.offset = 0;}
      oTL = oTRfound.beg;
      oDoc.TextSelection = new TextRange (oTL, oTL);
      oTRfound = KLD_F.FindSomething ("TXTTXT", sSearch, iMode, bCase, bWord, false);  // takes TextSelection
      return oTRfound;
    } else {
      return undefined;
    }

  } else if (sFType == "VARANY") { // ----------- Varaibles --------------------------------------- OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_VarBegin);
    oXX= aTi[0].obj;
    oTRfound = oXX.TextRange;
    return oTRfound;
  } else if (sFType == "VARNAM") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_VarBegin);
    for (j = 0; j < aTi.length; j++) {
      oXX = aTi[j].obj;
      if (KLD_F.SearchInString (oXX.VarFmt.Name, 0, sSearch, bCase, bWord, bRegEx) < 0) {
        continue;
      } else {
        return oXX.TextRange;
      }
    }
    return undefined;

  } else if (sFType == "XRFANY") { // ----------- Cross references -------------------------------- OK
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_XRefBegin);
    oXX= aTi[0].obj;
    oTRfound = oXX.TextRange;
    return oTRfound;
  } else if (sFType == "XRFFMT") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_XRefBegin);
    for (j = 0; j < aTi.length; j++) {
      oXX = aTi[j].obj;
      if (KLD_F.SearchInString (oXX.XRefFmt.Name, 0, sSearch, bCase, bWord, bRegEx) < 0) {
        continue;
      } else {
        return oXX.TextRange;
      }
    }
    return undefined;
  } else if (sFType == "XRFUNR") { // ------------
    aTi = oDoc.GetTextForRange (oTR, Constants.FTI_XRefBegin);
    for (j = 0; j < aTi.length; j++) {
      if (aTi[j].sData == undefined) {
        return aTi[j].obj.TextRange;
      } else {continue;}
    }
    return undefined;
  }

  if (aTi[0] == undefined) {return undefined;}
   oXX= aTi[0].obj;
  return KLD_F.ObjSelection (oXX);                // oXX.arFmt.Name
} // --- end FindInSelection ----------------------------------------------------------------------

KLD_F.FindLockedParagraph = function (oLastObj) { // ==============================================
/*            Find and identify locked paragraph
Arguments     oLastObj        Last object found:
                .obj          oPgf
                .row          unused
                .cell         unused
Returns       if found:       KLD_F.oLstObjFound updated
              else:           undefined
Called by     ButtonFind
Calling       DisplayObject
Used in       FMfindRepl
Reference     Own experiments
*/
var bFound = false, oDoc, oFlow, oFrame, oPgf, oTR;

  oDoc = KLD_F.oCurrentDoc;
	if (oLastObj == undefined || oLastObj.constructor.name != "Pgf") { // inital find
    if (oDoc.TextSelection.beg.obj.constructor.name == "Pgf") {
      oPgf = oDoc.TextSelection.beg.obj;
    } else {
      oFlow  = oDoc.MainFlowInDoc;
      oFrame = oFlow.FirstTextFrameInFlow;
      oPgf = oFrame.FirstPgf;
    }
	} else {
	  oPgf = oLastObj.NextPgfInFlow;            // advance
	}
  while (oPgf.ObjectValid()) {
    if (oPgf.Locked == true) {                    // find locked pgf
      bFound = true;
      oTR = new TextRange (
            new TextLoc (oPgf, 0),
            new TextLoc (oPgf, Constants.FV_OBJ_END_OFFSET));
      oDoc.TestSelection = oTR;
      oDoc.ScrollToText (oTR);
      oTR = oDoc.TextSelection;
      if (KLD_F.IsTextInset (oDoc, oTR)) {
        KLD_F.DisplayInfoF (KLD_F.UItxt.ButtonFind.@msg_16, "red");
        while (oPgf.Locked == true) {             // work off all ¶ in inset
            oPgf = oPgf.NextPgfInFlow;
         }
      }
      break;
    }
    oPgf = oPgf.NextPgfInFlow;
  }
  if (bFound) {
    KLD_F.oLstObjFound.object  = oPgf;
  } else {
    KLD_F.oLstObjFound.object  = undefined;
    KLD_F.oLstObjFound  = undefined;    // nothing found
  }
  return KLD_F.oLstObjFound;
} //--- end FindLockedParagraph -------------------------------------------------------------------

KLD_F.FindOverflowCell = function (oLastObj) { // =================================================
/*            Find and identify overflowing table cell
Arguments     oLastObj        Last object found:
                .obj          oTbl
                .row          oTblRow
                .cell         oTblCell
Returns       if found:       KLD_F.oLstObjFound updated
              else:           undefined
Called by     ButtonFind
Calling       DisplayObject
Used in       FMfindRepl
Reference     Rick Quatro, https://community.adobe.com/t5/framemaker/finding-cell-overflow-in-table-problem-with-script/m-p/7663483
              Klaus Göbel, https://community.adobe.com/t5/framemaker/view-current-cell-if-cell-overflows/m-p/11304625
History       2021-02-17 test with VSCode and ESDebugger
*/
var bFound = false, oDoc, oTbl, oTblRow, oTblCell, oTLbeg, oTLend;

  KLD_F.iError = 0;
  oDoc = KLD_F.oCurrentDoc;
  oTbl   = oLastObj.object;
  oTblRow = oLastObj.row;
  oTblCell= oLastObj.cell.NextCellInTbl;
out:
  while (oTbl.ObjectValid ()) {
    while (oTblRow.ObjectValid ()) {
      while (oTblCell.ObjectValid ()) {
        if (oTblCell.Overflowed === 1) {
          bFound = true;
          break out;
        }
        oTblCell = oTblCell.NextCellInRow;
      }
      oTblRow = oTblRow.NextRowInTbl;
      oTblCell = oTblRow.FirstCellInRow;
    }
    oTbl = oTbl.NextTblInDoc;
    oTblRow = oTbl.FirstRowInTbl;
    oTblCell = oTblRow.FirstCellInRow;
  }
  KLD_F.oLstObjFound.object  = oTbl;
  KLD_F.oLstObjFound.row     = oTblRow;
  KLD_F.oLstObjFound.cell    = oTblCell;

  if (bFound) {
    KLD_F.DisplayObject(oTblCell.FirstPgf);
    return KLD_F.oLstObjFound;
  } else {
    return undefined;                             // nothing found
  }
} //--- end FindOverflowCell ----------------------------------------------------------------------

KLD_F.FindOverflowText = function (oLastObj) { // =================================================
/*            Find and identify overflowing text flow
Arguments     -
Returns       -
Called by     ButtonFind
Calling       DisplayObject
Used in       FMfindRepl
Reference     Rick Quatro, https://community.adobe.com/t5/framemaker/find-overflow-text/td-p/4466878
History       2021-02-04 test with VSCode and ESDebugger
*/
var bFound = false, oDoc, oFlow, oTextFrame, oSubCol, oTLbeg, oTLend;

  oDoc = KLD_F.oCurrentDoc;
  oFlow   = oLastObj.object;
  KLD_F.iError = 0;

  while (oFlow.ObjectValid()) {                   // Get the last TextFrame and its last Sub Column.
    oTextFrame = oFlow.LastTextFrameInFlow;
    oSubCol = oTextFrame.LastSubCol;
    if (oSubCol.Overflowed === 1) {
        bFound = true;
        break;
    }
    oFlow = oFlow.NextFlowInDoc;
  }
  KLD_F.oLstObjFound.object  = oFlow;
  if (bFound) {
    KLD_F.DisplayObject(oTextFrame.LastPgf);
    oTextFrame.GraphicIsSelected = 1;
    return KLD_F.oLstObjFound;
  } else {
    return undefined;                             // nothing found
  }
} //--- end FindOverflowText ----------------------------------------------------------------------

KLD_F.FindSomething = function (sFType, sSearch, iMode, bWord, bCase, bBack) { // ==============
/*            Find anything defined by the parameters
Argument      sFType          Textual ID of Find Type. See Table asFindType
              sSearch         what to search for - used only for certain types
              iMode           0 Simple search
                              1 Wildcard search; excludes RegEx and simple Search
                              2 Regex search; excludes bWord and bBack; RegexFlavour is set to Perl
              bWord           true to constrain search to whole word
              bCase           true to consider case in search
              bBack           true to search backwards
Returns       true for found object, false if not found
Called by     FindInSelection, ButtonFind
Calling       GetFindParameters
Attention     Although the Find functions works for all types of objects, the end of the caller
              functions (ButtonFind) destroys the visible selection for the following objects:
              - Automatic Hyphen                  → Cursor is after the hyphen
              - Object Style Override             → No cursor visible
              - Object Style Tag                  → No cursor visible
Comment       NoWrap flag can not be used to stop search at end of document ...
              It is not possible to define a search which starts anywhere and ends at the visible
              end of the document. See reference.
              FA_errno          0: FE_Success
                               -2: FE_BadDocId
                              -43: FE_BadParameter
                              -65: FE_BadInsertPos
                              -41: FE_NotTextObject
                              -95: FE_NotFound
Reference     Russ Ward in https://forums.adobe.com/message/3888653#3888653
              Function Fno_FindNextEnXref to check for wrap around
              E:\_DDDprojects\_JSX-library\Docu\FMjsxLib.fm section Traversing a document
Used in       FMfindRepl
History       2020-12-04
*/
var oDoc = KLD_F.oCurrentDoc, oFindParms, oObject, oTL, oTR, oTRold, oTRsaved ;

  oTRsaved = oDoc.TextSelection;                  // we may need to come back to the start location
  oTR = oDoc.TextSelection;                       // start from current location
  oTRold = oTR;
  oFindParms = KLD_F.GetFindParameters (sFType, sSearch, iMode, bWord, bCase, bBack);
  if (oFindParms === null) {return undefined;}    // oTR has not changed yet
  oTR = oDoc.Find(oTR.beg, oFindParms);           // of searcjomg om Selection, result is after this
  KLD_F.iError = FA_errno;
  if (FA_errno == -95) {                          // nothing found (anymore)
    if (sFType== "CHRTAG") {
      return undefined;
    } else {
      return oTRold;
    }
  }
  if (!oTR.beg.obj.ObjectValid()) {               // object, not TR
    oObject = oDoc.FirstSelectedGraphicInDoc;
    if (oObject.ObjectValid()) {                  // convert o TR
      oTL = new TextLoc (oObject, 0)
      oTR = new TextRange (oTL, oTL)
    } else {
      oTR = oDoc.TextSelection;
    }
  }
// found TR / object is selected and visible; doc scrolls to location
  return oTR;
} //---  end FindSomething ------------------------------------------------------------------------

KLD_F.FindUserString = function (oDoc, aoUserStrings, sSearch, bCase, bWord, bRegEx) { // =========
/*            Find user string of certain text from certain object types
Arguments     oDoc            Current document
              aoUserStrings   Array with the collected objects
              sSearch         Text or RegEx to be found
              bCase           true to observe case in search
              bWord           true to check for full word
              bRegex          ture if sText is a RegEx
Global Var    aUserStrTypes   Array of names of objects from whic User Strings are searched.
                              ["AFrame", "Tbl"];
Returns       Object with the found string
                              object.ObjType  type of object
                              object.String   user string of object
                              object.TL       text location of object anchor
Called by     ButtonFind
Calling       SearchInString
Process       CollectUserStrings:
                In an array all objects with a non-empty user string are collected, together with TLoc
              FindUserString
                The find pokes the array and displays the type of object and the string
                Find next looks into the next item of the array...
Used in       FMfindRepl
History       2022-06-25
*/
var indx, j, j0 = KLD_F.jCurrentUStr+1, n = KLD_F.aoUserStrings.length, string;

  KLD_F.iError = 0;
  if (sSearch == undefined) {                     // search for empty user strings
    for (j = j0; j < n; j++) {
      string = KLD_F.aoUserStrings[j].Ustring;
      if (string == "" || string == undefined) {
        KLD_F.jCurrentUStr = j;
        return  null;                              // undefined, null, ""  trigger "not found"
      }
    }
  } else {
    for (j = j0; j < n; j++) {
      string = KLD_F.aoUserStrings[j].Ustring;
      indx = KLD_F.SearchInString(string, 0, sSearch, bCase, bWord, bRegEx);
      if (indx >= 0) {
        KLD_F.jCurrentUStr = j;
        return  KLD_F.aoUserStrings[j];
      }
    }
  }
  return undefined;                               // not found
} // --- end FindUserString -----------------------------------------------------------------------

KLD_F.GetBookComponents = function () { // ========================================================
/*            FIll array of book components (FM, MIF) from a book
Arguments     -
Returns       false in case of error
Called by     PalFindRepl, PrepareBookProc
Calling       DisplayInfoF, Message
Attention     Nested books are NOT considered, also XML, groupe and non-fm files
Comment       Generated components (TOC, IX etc.) are considered
              DFS order:  Depth-First search. Nested books are not entered.
Reference     ProcessBook in FMjsxLib
Used in       FMfindRepl
History       2022-08-09
*/
var sFileType, j, nFiles=0, oBook, oBookComp, oDoc, oEnDoc, oPfg, sDocName;
  oBook = app.ActiveBook;

  if (!oBook.ObjectValid()) {                     // No book is active
    KLD_F.DisplayInfoF (KLD_F.Message("e", KLD_F.UItxt.GetBookComponents.@msg_01), "red");
    return false;
  }
  KLD_F.aoBookComps.length = 0;                   // clear array
  KLD_F.bBookActive = true;
  oBookComp = oBook.FirstComponentInBook;
  while (oBookComp.ObjectValid()) {
    oBookComp.ComponentIsSelected = true;         // Indicate where we are --- debug only
    sFileType = oBookComp.BookComponentFileType;  // integer
    sName     = oBookComp.Name;
    if ((sFileType === Constants.FV_BK_FM )|| (sFileType === Constants.FV_BK_MIF)) {
      KLD_F.aoBookComps.push (new KLD_F.BookComponent(oBookComp, sFileType, sName));
    }
    oBookComp.ComponentIsSelected = false;        // --- debug only
    oBookComp = oBookComp.NextBookComponentInDFSOrder;
  }
  nFiles = KLD_F.aoBookComps.length;
  if (nFiles == 0) {                              // no FM, MIF files to be processed
    KLD_F.DisplayInfoF (KLD_F.Message("e", KLD_F.UItxt.GetBookComponents.@msg_02), "red");
    return false;
  }
  return true;
} //--- end GetBookComponents ---------------------------------------------------------------------

KLD_F.GetCharFmtsList = function (oDoc, asCharFmts) { //===========================================
/*            Fill array with list of character formats
Arguments     oDoc            Current document
              sArray          Array to be filled
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       -
History       2020-12-29
*/
var chrFormat = oDoc.FirstCharFmtInDoc;           // in Catalogue

  asCharFmts.length = 0;                          // clear the list
  while(chrFormat.ObjectValid() == true) {
    asCharFmts.push (chrFormat.Name);
    chrFormat = chrFormat.NextCharFmtInDoc;
  }
  asCharFmts.sort (KLD_F.SortNoCase);             // otherwise random order !
} //--- end GetCharFmtsList -----------------------------------------------------------------------

KLD_F.GetColourNameList = function (oDoc, aNames) { //=============================================
/*            Fill array with list of colours from the catalogue
Arguments     oDoc            Current document
              aNames          array to be filled with colour names: standard colours alphabetically,
                              then user colours alphabetically
                              A separator "------------" is inserted between these sublists
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       -
Comment       Formats are found in the catalogue in an unpredictable order
History       2021-02-09
*/
var j, oPage, oColour, pageType, sColourName, aoUsrColours = [], nUsrNames, aoSysColours = [], nSysNames;

  aNames.length = 0;                              // clear before refill
  oColour = oDoc.FirstColorInDoc;                 // variable in doc-catalogue
  while (oColour.ObjectValid()) {
    sColourName = oColour.Name;
    if (oColour.ReservedColor === Constants.FV_COLOR_NOT_RESERVED) { // user colour ?
      aoUsrColours.push (sColourName);
    } else {                                      // prefined colour
      aoSysColours.push (sColourName);
    }
    oColour = oColour.NextColorInDoc;
  }
  aoUsrColours.sort (KLD_F.SortNoCase);           //  otherwise random order !
  aoSysColours.sort (KLD_F.SortNoCase);           // #, 1, A, a, …
  nUsrNames = aoUsrColours.length;
  for (j = 0; j < nUsrNames; j++){                // add user colours
    aNames.push(aoUsrColours[j]);
  }
  aNames.push(KLD_F.UItxt.GetColourNameList.@sep_01);// insert separator line
  nSysNames = aoSysColours.length;
  for (j = 0; j < nSysNames; j++){                // add system colours
    aNames.push(aoSysColours[j]);
  }
  return nUsrNames;
} //--- end GetColourNameList ---------------------------------------------------------------------

KLD_F.GetCondFmtsList = function (oDoc, asCondNames) { //==========================================
/*            Fill array with list of condition names
Arguments     oDoc            Current document
              asCondNames     Array getting the condition names
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       IsOneOf
Comment       Formats are found in the document in an unpredictable order
              Until FM-16 conditions for system use are listed in the conditions catalogue
              - they should be hidden from the user.
Reference     xxx
Used in       FMfindRepl
History       2020-12-09
              2022-06-22 do not put FM_PDF_xxx items into the list
*/
var oCondFmt,
    asToHide = "FM8_SYSTEM_HIDEELEMENT FM8_TRACK_CHANGES_ADDED FM8_TRACK_CHANGES_DELETED FM_PDF_Comments_Highlight FM_PDF_Comments_Underline";

  oCondFmt = oDoc.FirstCondFmtInDoc
  asCondNames.length = 0;                         // clear the list
  while(oCondFmt.ObjectValid() == true) {
    if (KLD_F.IsOneOf (asToHide, oCondFmt.Name) > 0) { // skip the system used items
      oCondFmt = oCondFmt.NextCondFmtInDoc;
      continue;
    } else {
      asCondNames.push (oCondFmt.Name);
      oCondFmt = oCondFmt.NextCondFmtInDoc;
    }
  }
  asCondNames.sort (KLD_F.SortNoCase);            // otherwise random order !
} //--- end GetCondFmtsList -----------------------------------------------------------------------

KLD_F.GetFindParameters = function (sFType, sSearch, iMode, bWord, bCase, bBack) { //==============
/*            Set up search parameters accordingly
Argument      sFType          Textual ID of Find Type. See Table asFindType
              sSearch         what to search for - used only for certain types
              iMode           0 Simple search
                              1 Wildcard search; excludes RegEx and simple Search
                              2 Regex search; excludes bWord and bBack; RegexFlavour is set to Perl
              bWord           true to constrain search to whole word
              bCase           true to consider case in search
              bBack           true to search backwards
Returns       Parameters for the find method
Comment       FS_FindWrap: A flag that determines whether the find operation wraps when it
              reaches the location where the search began.
              Default is True; the find operations wraps.
              If False, after reaching the location where the search began, the find operation
              returns an empty TextRange and  FA_errno is set to FE_NotFound (-95).
              → This flag can not be used to stop search at end of document ...
              The following Constants.FS_xx are not handled in this routine:
               2 FS_FindElementTag
               7 not defined
              14 FS_FindCondTextNotInCondTags
Called by     CollectUserStrings, FindSomething
Calling       DisplayInfoF
Reference     FDK Function Reference, F_ApiFind(), https://forums.adobe.com/thread/961616
Used in       FMfindRepl
History       2020-12-01
*/
var bWild, bRegEx, findParms, jj, qFlags, oPropVal1, oPropVal2, oPropVal3, oPropVal4;

  if (iMode == undefined) {iMode = 0;}
  if (bWord == undefined) {bWord = false;}
  if (bCase == undefined) {bCase = false;}
  if (bBack == undefined) {bBack = false;}  bWild  = iMode == 1;
  bRegEx = iMode == 2;

  findParms = new PropVals();
// This setting has (in most cases) no effect. See function IsWrapped -----------------------------
  oPropVal1 = new PropVal() ;
  oPropVal1.propIdent.num = Constants.FS_FindWrap ;
  oPropVal1.propVal.valType = Constants.FT_Integer;
  oPropVal1.propVal.ival = 0 ;                    // no wrap
  findParms.push(oPropVal1);
//
  oPropVal2 = new PropVal() ;
  if        (sFType == "ANCFRM") { // ----- Constants.FS_FindObject, Constants.FV_FindAnchoredFrame
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAnchoredFrame ;
  } else if (sFType == "AUTHYP") { // --- Constants.FS_FindObject, Constants.FV_FindAutomaticHyphen
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAutomaticHyphen ;
  } else if (sFType == "CHRFMO") { // Constants.FS_FindObject, Constants.FV_FindCharacterFormatOverride
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindCharacterFormatOverride ;
  } else if (sFType == "CHRFMT") { // ------------------------------------ Constants.FS_FindCharFmt
    // this should not appear in the call
    KLD_F.DisplayInfoF (KLD_F.Message ("e", KLD_F.UItxt.GetFindParameters.@msg_02, "GetFindParameters", sFType), "red");
    return null;
  } else if (sFType == "CHRTAG") { // ------------------------------------ Constants.FS_FindCharTag
    oPropVal2.propIdent.num = Constants.FS_FindCharTag;
    oPropVal2.propVal.valType = Constants.FO_CharFmt;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "CONTAG") { // ------------------------ Constants.FS_FindCondTextInCondTags
    oPropVal2.propIdent.num = Constants.FS_FindCondTextInCondTags;
    oPropVal2.propVal.valType = Constants.FT_Strings;
    oPropVal2.propVal.ssval[0] = sSearch;         // "cond1"
//  for (jj = 0; jj < sSearch.length; jj++) {     // ["cond1", "cond2", …] does not work in ES
//    oPropVal2.propVal.ssval[jj] = sSearch[jj];
//  }
  } else if (sFType == "FOONOT") { // ---------- Constants.FS_FindObject, Constants.FV_FindFootnote
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindFootnote ;
  } else if (sFType == "MRKANY") { // --------- Constants.FS_FindObject, Constants.FV_FindAnyMarker
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAnyMarker ;
  } else if (sFType == "MRKTXT") { // -------------------------------- Constants.FS_FindMarkerText
    oPropVal2.propIdent.num = Constants.FS_FindMarkerText;
    oPropVal2.propVal.valType = Constants.FT_String;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "MRKTYP") { // ------------------------------ Constants.FS_FindMarkerOfType
    oPropVal2.propIdent.num = Constants.FS_FindMarkerOfType;
    oPropVal2.propVal.valType = Constants.FT_String;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "OSTFMO") { // Constants.FS_FindObject, Constants.FV_FindStyleFormatOverride
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindStyleFormatOverride ;
  } else if (sFType == "OSTTAG") { // ----------------------------------- Constants.FS_FindStyleTag
    oPropVal2.propIdent.num = Constants.FS_FindStyleTag;
    oPropVal2.propVal.valType = Constants.FT_String;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "PGFFMO") { // - Constants.FS_FindObject, Constants.FV_FindPgfFormatOverride
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindPgfFormatOverride ;
  } else if (sFType == "PGFTAG") { // ------------------------------------- Constants.FS_FindPgfTag
    oPropVal2.propIdent.num = Constants.FS_FindPgfTag;
    oPropVal2.propVal.valType = Constants.FT_String;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "RUBANY") { // ----------- Constants.FS_FindObject, Constants.FV_FindAnyRubi
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAnyRubi ;
  } else if (sFType == "TBLANY") { // ---------- Constants.FS_FindObject, Constants.FV_FindAnyTable
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAnyTable ;
  } else if (sFType == "TBLFMO") { // Constants.FS_FindObject, Constants.FV_FindTableFormatOverride
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindTableFormatOverride ;
  } else if (sFType == "TBLTAG") { // ----------------------------------- Constants.FS_FindTableTag
    oPropVal2.propIdent.num = Constants.FS_FindTableTag;
    oPropVal2.propVal.valType = Constants.FO_TblFmt;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "TINANY") { // ------ Constants.FS_FindObject, Constants.FV_FindAnyTextInset
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAnyTextInset ;
  } else if (sFType == "TINUNR") { // Constants.FS_FindObject, Constants.FV_FindUnresolvedTextInset
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindUnresolvedTextInset ;
  } else if (sFType == "TXTCLP") { // ---------------------------------- Constants.FS_FindSelection
//  oPropVal2.propIdent.num = Constants.FS_FindSelection;
//  oPropVal2.propVal.valType = Constants.FT_TextRange;
//  oPropVal2.propVal.trval = sSearch;            // this is oTR - but how to get it from the clipboard?
    KLD_F.DisplayInfoF (KLD_F.Message ("e", KLD_F.UItxt.GetFindParameters.@msg_03, "GetFindParameters", sFType), "red");
    return null;
  } else if (sFType == "TXTTXT") { // --------------------------------------- Constants.FS_FindText
    oPropVal2.propIdent.num = Constants.FS_FindText;
    oPropVal2.propVal.valType = Constants.FT_String;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "VARANY") { // ------- Constants.FS_FindObject, Constants.FV_FindAnyVariable
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAnyVariable ;
  } else if (sFType == "VARNAM") { // ----------------------------- Constants.FS_FindNamedVariable
    oPropVal2.propIdent.num = Constants.FS_FindNamedVariable;
    oPropVal2.propVal.valType = Constants.FT_String;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "XRFANY") { // ------------- Constants.FS_FindObject, Constants.FV_FindAnyXRef
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindAnyXRef ;
  } else if (sFType == "XRFFMT") { // ----------------------------- Constants.FS_FindXRefWithFormat
    oPropVal2.propIdent.num = Constants.FS_FindXRefWithFormat;
    oPropVal2.propVal.valType = Constants.FT_String;
    oPropVal2.propVal.sval = sSearch;
  } else if (sFType == "XRFUNR") { // ---- Constants.FS_FindObject, Constants.FV_FindUnresolvedXRef
    oPropVal2.propIdent.num   = Constants.FS_FindObject;
    oPropVal2.propVal.valType = Constants.FT_Integer;
    oPropVal2.propVal.ival = Constants.FV_FindUnresolvedXRef ;
  } else {
    KLD_F.DisplayInfoF (KLD_F.Message ("e", KLD_F.UItxt.GetFindParameters.@msg_01, "GetFindParameters", sFType), "red");
    return null;
  }
  findParms.push(oPropVal2);

  if (bWord || bCase || bBack || bRegEx || bWild) { // customisation flags are ORED
    oPropVal3 = new PropVal() ;
    oPropVal3.propIdent.num = Constants.FS_FindCustomizationFlags ;
    oPropVal3.propVal.valType = Constants.FT_Integer;
    qFlags = 0;
    if (bCase)  { qFlags = qFlags   | Constants.FF_FIND_CONSIDER_CASE;} // 0x01
    if (bRegEx) { qFlags = qFlags   | Constants.FF_FIND_USE_REGEX;      // 0x10
        oPropVal4 = new PropVal ();
        oPropVal4.propIdent.num = Constants.FS_RegexFlavour;            // 17
        oPropVal4.propVal.valType = Constants.FT_Integer;
        oPropVal4.propVal.ival = Constants.FR_USE_PERL;                 //  1
        findParms.push(oPropVal4);
    } else {
      if (bWild)  { qFlags = qFlags | Constants.FF_FIND_USE_WILDCARDS;} // 0x04
      if (bWord)  { qFlags = qFlags | Constants.FF_FIND_WHOLE_WORD;}    // 0x02
      if (bBack)  { qFlags = qFlags | Constants.FF_FIND_BACKWARDS;}     // 0x08
    }
    oPropVal3.propVal.ival = qFlags;              // ival reports error in ESTK
    findParms.push(oPropVal3);
  }
return findParms;
} //---  end GetFindParameters --------------------------------------------------------------------

KLD_F.GetMarkerTypeList = function (oDoc, asMarkerTypes) { //======================================
/*            Fill array with list of Marker type names
Arguments     oDoc            Current document
              asMarkerTypes   Array to be filled
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       -
Comment       Formats are found in the catalogue in an unpredictable order
History       2020-12-29
*/
var sMarkerType;

  sMarkerType = oDoc.FirstMarkerTypeInDoc;        // in Catalogue
  asMarkerTypes.length = 0;                       // clear the list
  while(sMarkerType.ObjectValid() == true) {
    asMarkerTypes.push (sMarkerType.Name);
    sMarkerType = sMarkerType.NextMarkerTypeInDoc;
  }
  asMarkerTypes.sort (KLD_F.SortNoCase);          // otherwise random order !
} //--- end GetMarkerTypeList ---------------------------------------------------------------------

KLD_F.GetPage = function (oDoc, obj) { //===
/*            Get the page with the object on it ==================================================
Arguments     obj:  object (variable, ...) for which to find the page
              oDoc: document with the object
Returns       Function value =  index object oPage
Called by     DisplayObject, CollectUserStrings
Calling       -
Comment       The graphic object math must be written with lower case initital
              FDK lists FO_Math, ESTK lists Math, but an object inspection reveils it to be math
Reference     GetPage by Rick Quatro, https://forums.adobe.com/thread/2424187
Used in       FMvars, FMfindRepl
History       2020-12-01
*/
var oFrame = 0, cell = 0, objType = "", oProp = 0;

  while (obj) {
    oFrame = obj;
    objType = obj.constructor.name;
    switch (objType) {
      case "SubCol" :
        obj = obj.ParentTextFrame;
        break;
      case "Tbl" :
        obj = obj.FirstRowInTbl.FirstCellInRow;
        break;
      case "Row" :
        obj = obj.FirstCellInRow;
        break;
      case "Cell" :
      case "Pgf" :
      case "AFrame" :
        obj = obj.InTextFrame;
        break;
      case "TextLine" :
      case "TextFrame" :
      case "UnanchoredFrame" :
      case "Arc" :
      case "Ellipse" :
      case "Group" :
      case "Inset" :
      case "Line" :
      case "math" :
      case "Polygon" :
      case "Polyline" :
      case "Rectangle" :
      case "RoundRect" :
        if (obj.FrameParent.ObjectValid()) {
            obj = obj.FrameParent;
        } else {
            obj = 0;
        }
        break;
      case "XRef" :
      case "Var" :
        oProp = oDoc.GetTextPropVal (obj.TextRange.beg, Constants.FP_InTextObj);
        obj = oProp.propVal.obj;
        break;
      default:                                    // Prevent endless loop if unknown object type is found.
        obj = 0;
        break;
    }
  }
  if (oFrame) {
    return oFrame.PageFramePage;
  } else {
    return 0;
  }
} //--- end GetPage

KLD_F.GetPgfFmtsList = function (oDoc, asPfgFmts) { //=============================================
/*            Fill array with list of paragraph formats
Arguments     oDoc            Current document
              asPfgFmts       Array to be filled
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       -
Comment       Formats are found in the catalogue in an unpredictable order
History       2020-12-29
*/
var pgfFormat = oDoc.FirstPgfFmtInDoc;            // in Catalogue

  asPfgFmts.length = 0;                           // clear the list
  while(pgfFormat.ObjectValid() == true) {
    asPfgFmts.push (pgfFormat.Name);
    pgfFormat = pgfFormat.NextPgfFmtInDoc;
  }
  asPfgFmts.sort (KLD_F.SortNoCase);              // otherwise random order !
} //--- end GetPgfFmtsList ------------------------------------------------------------------------

KLD_F.GetNextObjectInDoc = function (obj) { // ====================================================
/*            Get the next object of the given type
Arguments     obj             Object for which to get the next one. The following types are handled:
                              Graphic objects (see variable aGraphicObj), Pgf, Cell.
Returns       next object of same type | undefined
Called by     FindColourUse, ButtonFind
Calling       IsGraphicObject
Attention     For Cell objects this functions looks in the document for the next cell,
              not just in the current table! Row objects stay within the current table.
Comment       To be independant, IsGraphicObject is integrated here.
              Concerning the order of handling the objects:
              nPary ≥ nTable + nCells; nGraphics can be anything between 0 and a large number.
              The graphic object math must be written with lower case initital
              FDK lists FO_Math, ESTK lists Math, but an object inspection reveils it to be math
Reference     -
Used in       FMfindRepl
History       2021-02-25
*/
var oRow, oTbl;

  if (obj.constructor.name == "Pgf") {
    obj = obj.NextPgfInDoc;
    if (obj.ObjectValid()) KLD_F.oLstObjFound.object = obj;
  } else if (obj.constructor.name == "Cell") {
    oRow = obj.CellRow;
    oTbl = oRow.RowTbl.NextTblInDoc;              // look for next table prophylactivally
    if (obj.NextCellInTbl.ObjectValid()) {
      obj = obj.NextCellInTbl;
    } else {                                      // need to go to next table
      if (oTbl.ObjectValid()) {
        obj = oTbl.FirstRowInTbl.FirstCellInRow;
      } else {                                    // no more tables
        obj = undefined;
      }
    }
    if (obj != undefined) KLD_F.oLstObjFound.object = obj;
  } else if (KLD_F.IsGraphicObject(obj)) {
    obj = obj.NextGraphicInDoc;
    if (obj.ObjectValid()) KLD_F.oLstObjFound.object = obj;
  }
  if (obj != undefined && obj.ObjectValid()) {
    return KLD_F.oLstObjFound.object;
  } else {
    return undefined;
  }
} //--- end GetNextObjectInDoc --------------------------------------------------------------------

KLD_F.GetObjFmtList = function (oDoc, asObjFmts) { //=== Fill array with list of Marker type names
/*            Fill array with list of Object format names
Arguments     oDoc            Current document
              asObjFmts       Array to be filled with the Graphic object style names
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       -
Comment       Formats are found in the catalogue in an unpredictable order
History       2020-12-29
*/
var sObjFmt;                                      // in Catalogue

sObjFmt = oDoc.FirstGraphicsFmtInDoc;
  asObjFmts.length = 0;                           // clear the list
  while(sObjFmt.ObjectValid() == true) {
    asObjFmts.push (sObjFmt.StyleTag);
    sObjFmt = sObjFmt.NextGraphicsFmtInDoc;
  }
  asObjFmts.sort (KLD_F.SortNoCase);              // otherwise random order !
} //--- end GetObjFmtList -------------------------------------------------------------------------

KLD_F.GetTableFmtList = function (oDoc, asTableFmts) { //==========================================
/*            Fill array with list of Table formats
Arguments     oDoc            Current document
              asObjFmts       Array to be filled with the Graphic object style names
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       xxx
Comment       Formats are found in the catalogue in an unpredictable order
History       2020-12-29
*/
var sTableFmt;                                    // in Catalogue

  sTableFmt = oDoc.FirstTblFmtInDoc;
  asTableFmts.length = 0;                         // clear the list
  while(sTableFmt.ObjectValid() == true) {
    asTableFmts.push (sTableFmt.TblTag);
    sTableFmt = sTableFmt.NextTblFmtInDoc;
  }
  asTableFmts.sort (KLD_F.SortNoCase);            // otherwise random order !
} //--- end GetTableFmtList -----------------------------------------------------------------------

KLD_F.GetTextOfSelection = function () { // =======================================================
/*            Get the text from the current selection
Returns       Textstring in case the selection is within a paragraph
Called by     PalFindRepl, ButtonReplace
Calling       -
History       2022-09-09
*/
var aTextItems, oDoc, oTR, sTxt;

  oDoc = app.ActiveDoc;
  oTR  = oDoc.TextSelection;
  if (oTR.beg.obj.constructor.name == "Pgf") {
    aTextItems = oDoc.GetTextForRange (oTR, Constants.FTI_String);
    if (aTextItems[0] == undefined) {return "";}  // no text selection
    sTxt = sText = (aTextItems[0].sdata);
    return sTxt;
  }
} //--- end GetTextOfSelection -----------------------------------------------------------------------

KLD_F.GetVarNameList = function (oDoc, aNames, aoSourceVars) { //==================================
/*            Fill array with list of variables from the catalogue
Arguments     oDoc            Current document
              aNames          array to be filled with var names: user vars alphabetically,
                              then system vars alphabetically
                              A separator "------------" is inserted after the user vars
              aoSourceVars    array of variable objects (name, value)
Returns       aNames, aoSourceVars
Called by     GetCatalogAny, GetCatalogF
Calling       Variable, SortNoCase2
Comment       Formats are found in the catalogue in an unpredictable order
              Since both name and content of the variables are needed for transfer,
              array aoSourceVars is set up in parallel
History       2020-12-29
*/
var j, oPage, oVar, pageType, sVarName, aoUsrVars = [], nUsrNames, aoSysVars = [], nSysNames;

  aNames.length = 0;                              // clear before refill
  aoSourceVars.length = 0;
  oVar = oDoc.FirstVarFmtInDoc;                   // variable in doc-catalogue
  while (oVar.ObjectValid()) {
    sVarName = oVar.Name;
    sVarValue = oVar.Fmt;
    if (oVar.SystemVar === Constants.FV_VAR_USER_VARIABLE) { // user variable ?
      aoUsrVars.push (new KLD_F.Variable(sVarName, sVarValue));
    } else {                                      // system variable
      aoSysVars.push (new KLD_F.Variable(sVarName, sVarValue));
    }
    oVar = oVar.NextVarFmtInDoc;
  }
  aoUsrVars.sort (KLD_F.SortNoCase2);             //  otherwise random order !
  aoSysVars.sort (KLD_F.SortNoCase2);             // #, 1, A, a, …
  nUsrNames = aoUsrVars.length;
  for (j = 0; j < nUsrNames; j++){                // add user vars
    aoSourceVars.push(aoUsrVars[j]);
    aNames.push(aoUsrVars[j].Name);
  }
  aNames.push(KLD_F.UItxt.GetVarNameList.@sep_01);// insert separator line
  aoSourceVars.push("");                          // keep it in sync
  nSysNames = aoSysVars.length;
  for (j = 0; j < nSysNames; j++){                // add system vars
    aoSourceVars.push(aoSysVars[j]);
    aNames.push(aoSysVars[j].Name);
  }
  return nUsrNames;
} //--- end GetVarNameList ------------------------------------------------------------------------

KLD_F.GetXrefFmtsList = function (oDoc, asXrefNames) { //==========================================
/*            Fill array with list of x-ref formats
Arguments     oDoc            Current document
              asXrefNames     Array to be filled
Returns       -
Called by     GetCatalogAny, GetCatalogF
Calling       -
Comment       Formats are found in the catalogue in an unpredictable order
History       2020-12-29
*/
var xrefFormat = oDoc.FirstXRefFmtInDoc;          // in Catalogue

  asXrefNames.length = 0;                         // clear the list
  while(xrefFormat.ObjectValid() == true) {
    asXrefNames.push (xrefFormat.Name);
    xrefFormat = xrefFormat.NextXRefFmtInDoc;
  }
  asXrefNames.sort (KLD_F.SortNoCase);            // otherwise random order !
} //--- end GetPgfFmtsList ------------------------------------------------------------------------

KLD_F.IsTextInset = function (oDoc, oTRold) {// ============================================
/*            Check whether the Text Range is a Text Inset
Arguments     oDoc            Current document
              oTRold          TR of the assumed inset
Returns       True if the TR  constitutes a Text Inset
Called by     FindLockedParagraph
Calling       -
Comment       I have not found another  method to identify the TR as TI than to search for it
History       2022-12-31
*/
var aTextItems, iTxtLen;

  if (oTRold.constructor.name !== "TextRange") {
    aTextItems = oTRold.GetText(Constants.FTI_TextInsetBegin);               // text objects
  } else {
    aTextItems = oDoc.GetTextForRange(oTRold, Constants.FTI_TextInsetBegin); // text range
  }
  iTxtLen = aTextItems.len;

  if (iTxtLen > 0) {                              // Text Inset found
    return true;
  } else {
    return false;
  }
} //--- end IsTextInset ---------------------------------------------------------------------------

KLD_F.IsWrapped = function (oCurrentFindTR) { // ==================================================
/*            Check whether the find process has wrapped around
Arguments     oCurrentFindTR  TR of the current find cycle
Globals       bFirstFind, oFirstFindTR
Returns       True if the starting TL has been reached again - wrapping happened, else False
Called by     ButtonFind
Calling       -
Comment       If there is only one occurrence of the find item, then only at the second Find
              the wrap will be detected.
Reference     https://community.adobe.com/t5/framemaker-discussions/to-wrap-or-not-to-wrap-that-is-the-question/m-p/13147485#M76017
History       2022-08-22
*/
  if (KLD_F.bFirstFind) {
    KLD_F.oFirstFindTR = oCurrentFindTR;
    KLD_F.bFirstFind   = false;
  } else {
    if (oCurrentFindTR.beg.obj.Unique == KLD_F.oFirstFindTR.beg.obj.Unique &&
        oCurrentFindTR.beg.offset == KLD_F.oFirstFindTR.beg.offset) {
      return true;
    }
  }
  return false;
} //--- end IsWrapped -----------------------------------------------------------------------------

KLD_F.ListSelector = function (oEvt) { // =========================================================
/*            KB event handler for keydown. There is no event keypress defined in ES
Arguments     oEvt            Event object to be handled.
Globals       KLD_F.evBuffer  Takes the sequence of typed keys, Backspace removes last entered
              evParm          Must be defined in the target (The element object where the event occurres)
                              This defines the array of items in the list.
Returns       oEvt.currentTarget.selection is set directly
Called by     List controls in PalFindRepl
Comment       • If the list provided is not in alphabetical order it will be difficult to find
                some items: typing k find Kastanien. But you have no immediate knowledge about
                Kartoffeln. you need to know that it exists and type Kar to select it.
              • With up/down arrow keys you can further change the selection.
              • After navigation Enter must be used to execute the corresponding catatalgoue function
Attention     This function can not handle characters created with modifyer keys. For example
              on a Swiss German keyboard you nedd to first type ALTGR+´ and then SHIFT+e to
              create the character É - I assumed that this can be found by typing é. But this
              é is SHIFT+ö — I have no clue how to handle this.
Reference     Peter Kahrel, "SciptUI for Dummies": Selecting items in dropdowns using the keyboard
              https://stackoverflow.com/questions/256754/how-to-pass-arguments-to-addeventlistener-listener-function
Used in       FMfindRepl
History       2022-09-12
*/
var array, j=0, kName, kChar, kID, sFunction;
  array  = oEvt.currentTarget.evParm;
  kName = oEvt.keyName.toLowerCase();
  if (kName == "down" || kName == "up") {return}
  if (kName == "enter"|| kName == "tab") {        // leave the navigation cycle
    KLD_F.evBuffer = "";
    return;
  }
  if (kName == "backspace") {
    KLD_F.evBuffer = KLD_F.evBuffer.replace (/.$/, "");
  } else {
    kID = oEvt.keyIdentifier;                     // U+00FC → ü
    kID = kID.substring  (2);                     //   00FC
    kChar = parseInt(kID, 16);                    //    252
    kChar = String.fromCharCode(kChar);           //      ü
    kChar = kChar.toLowerCase();
    KLD_F.evBuffer += kChar;
    while (j < array.length-1 && array[j].toLowerCase().indexOf (KLD_F.evBuffer) != 0) {
      ++j;
    }
  }
  if (array[j].toLowerCase().indexOf (KLD_F.evBuffer) == -1) {
    oEvt.currentTarget.selection = 0;
  } else {
    oEvt.currentTarget.selection = j;
  }
} // --- end ListSelector -------------------------------------------------------------------------

KLD_F.PrepareBookProc = function () { // ==========================================================
/*            Prepare book processing
Arguments     -
Called by     Notify, PalFindRepl
Calling       DisplayInfoF, GetBookComponents, ClearInfoArea
History       2022-08-23
*/
  KLD_F.oCurrentBook = app.ActiveBook;
  if (!KLD_F.oCurrentBook.ObjectValid()) {
    KLD_F.DisplayInfoF (KLD_F.UItxt.General.@BookRequired, "red");
    wPalF.g07.rbBook.value = false;               // leave it off
    return;
  }
  KLD_F.bDocActive = false;
  KLD_F.bBookActive = true;
  KLD_F.GetBookComponents();
  KLD_F.kCurrentComp = -1;                        // we are starting search in book
  KLD_F.bNextCompOK = true;                       // to get the first one
  KLD_F.ClearInfoArea ();
} //--- end IsWrapped -----------------------------------------------------------------------------

KLD_F.RemoveCondFmtFromRow = function (oDoc, oRow, sCondFmt) { // =================================
/*            Remove selected condition format
Arguments     oDoc            Current document
              oRow            Table row with the condition
              sCondFmt        Condition format (name) to be removed
Called by     ButtonReplace
Calling       -
Reference     http://frameautomation.com/category/extendscript/page/3/
              https://community.adobe.com/t5/framemaker-discussions/replace-condition-format/td-p/13096681
History       2022-07-28
*/
var j, aoCondFmts;

  aoCondFmts = oRow.InCond;                       // Get list of the conditions applied to the row.
  if (aoCondFmts.length === 0) {
    return;                                       // No conditions are applied to the row.
  }
  for (j = 0; j < aoCondFmts.length; j += 1) {    // Loop through the format names in the row.
    if (aoCondFmts[j].Name == sCondFmt) {         // Should format  be removed?
      Array.prototype.splice.call (aoCondFmts, j, 1); // Remove it from the list.
    }
  }
  if (aoCondFmts.length !== oRow.InCond.length) {
    oRow.InCond = aoCondFmts;                     // Conditions were removed; apply the updated list
  }
} //--- end RemoveCondFmtFromRow ------------------------------------------------------------------

KLD_F.RemoveConditionsFromText = function (oDoc, oTR, sCondFmt) { // ==============================
/*            Remove selected condition format
Arguments     oDoc            Current document
              oTR             Text range with the condition
              sCondFmt        Condition format (name) to be removed
Returns       Modified Text range
Called by     ButtonReplace
Calling       -
Reference     http://frameautomation.com/category/extendscript/page/3/
              https://community.adobe.com/t5/framemaker-discussions/replace-condition-format/td-p/13096681
History       2022-07-28
*/
var j, n, aoCondFmts = [], oCondFmt, oProps;

  oProps = oDoc.GetTextPropVal (oTR.beg, Constants.FP_InCond); // List of the conditions applied
  n = oProps.propVal.osval.length;
  if (n == 0) {
      return;                                     // No conditions are applied to the text range
  }
  for (j = 0; j < n; j += 1) {                    // Loop through the applied condition format names
    if (oProps.propVal.osval[j].Name == sCondFmt) { // Should format  be removed?
      Array.prototype.splice.call (oProps.propVal.osval, j, 1); // Both osval and isval must be
      Array.prototype.splice.call (oProps.propVal.isval, j, 1); // removed from the list.
    }
  }
  if (oProps.propVal.osval.length != n) {         // oProps have been changed
    oDoc.SetTextPropVal (oTR, oProps);
  }
} //--- end RemoveConditionsFromText --------------------------------------------------------------

KLD_F.RemoveOverride = function (oDoc, sFType) { // ===============================================
/*            Remove overrides from various objects
Arguments     oDoc            Current document
              sFType          Object type defined by KLD's abbreviation
Returns       -
Called by     ButtonReplace
Calling       AdvanceTextLoc, ExecuteFCodes, DisplayInfoF, ApplyObjectStyle, ApplyPgfFmt, ApplyTblFmt
Used in       FMfindRepl
History       2022-06-22
*/
var abc;

var aItems = [], j, n, oDoc = KLD_F.oCurrentDoc,
    oTR, oObject, oParent, oPgf, oTbl, oTL, oTR, sPgfFmt, sObjFmt, sStyleTag, sTblFmt;

oTR = oDoc.TextSelection;
  if (sFType == "CHRFMO") {                       // --- Character overrides ----------------------
    KLD_F.ExecuteFCodes (["CharacterDefaultPgfFont"]); // remove the overwrite¶
    oDoc.TextSelection = oTR;                     // needs to be done twice
    KLD_F.ExecuteFCodes (["CharacterDefaultPgfFont"]);

  } else if (sFType == "OSTFMO") {                // --- Object Style  overrides ------------------
    oObject = oDoc.TextSelection;
    KLD_F.DisplayInfoF (KLD_F.UItxt.RemoveOverride.@msg_01, "red");

    if (oObject.beg.obj.constructor.name == "Pgf") {
      oObject = oObject.beg.obj;
      aItems =  oObject.GetText (Constants.FTI_FrameAnchor);  // text objects
      n = aItems.length;
      for (j = 0; j < n; j++) {
        oObject = aItems[j].obj;
        sStyleTag = oObject.StyleTag;
        KLD_F.ApplyObjectStyle (oDoc, oObject, sStyleTag);// Re-apply style
      }
    } else {
      oObject = oDoc.FirstSelectedGraphicInDoc;
      sStyleTag = oObject.StyleTag;
      KLD_F.ApplyObjectStyle (oDoc, oObject, sStyleTag); // Re-apply style
      oParent = oObject.FrameParent;
      KLD_F.AdvanceTextLoc (oParent);
    }

  } else if (sFType == "PGFFMO") {                // --- Paragraph format overrides ---------------
    oPgf = oTR.beg.obj;
    sPgfFmt = oPgf.Name;
    KLD_F.ApplyPgfFmt (oDoc, oPgf, sPgfFmt);      // Complete format must be applied

  } else if (sFType == "TBLFMO") {                // --- Table format overrides -------------------
    oObject = oDoc.TextSelection;
    if (oObject.beg.obj.constructor.name == "Pgf") {
      oObject = oObject.beg.obj;
      aItems =  oObject.GetText (Constants.FTI_TblAnchor);  // text objects
      n = aItems.length;
      for (j = 0; j < n; j++) {
        oTbl = aItems[j].obj;
        sTblFmt = oTbl.TblTag;
        KLD_F.ApplyTblFmt (oDoc, oTbl, sTblFmt);  // Complete format must be applied
      }
    }
    oTR = new TextRange(oTbl.TextLoc, oTbl.TextLoc); // advance needed?
    oDoc.TextSelection = oTR;
  }
} //--- end RemoveOverride ------------------------------------------------------------------------

KLD_F.WarnBackwardsSearch = function () { // ======================================================
/*            Certain Find Types may cause FM to hang when serching backwards
Called by     PalFindRepl
Calling       IsOneOf, DisplayInfoF
Comment       This is not specific to FMfindRepl. It may also happen to the standard FM process
              See documentation FMfindRepl
Reference     Own experiments
Used in       FMfindRepl
History       2022-06-20
*/
var iFT, sFname, sFType;
  iFT       = wPalF.g01.lstFindTypeF.selection.index;
  sFType    = KLD_F.asFType[iFT];
  sFname    = wPalF.g01.lstFindTypeF.selection;
  if (KLD_F.IsOneOf ("MRKANY MRKTXT MRKTYP OSTTAG RUBANY TBLFMO TBLANY TBLTAG VARANY VARNAM XRFANY XRFFMT XRFUNR ", sFType) > 0) {
   KLD_F.DisplayInfoF (KLD_F.Message ("e", KLD_F.UItxt.wPalF.@g04chkBackWarn, sFname), "red");
  }
} //--- end WarnBackwardsSearch -------------------------------------------------------------------
